"""TriangulatedPatch class used by Surface class."""

import logging

log = logging.getLogger(__name__)

import numpy as np

import resqpy.olio.uuid as bu
import resqpy.olio.vector_utilities as vec
import resqpy.olio.xml_et as rqet


class TriangulatedPatch:
    """Class for RESQML TrianglePatch objects (used by Surface objects inter alia)."""

    def __init__(self, parent_model, patch_index = None, patch_node = None, crs_uuid = None):
        """Create an empty TriangulatedPatch (TrianglePatch) node and optionally load from xml.

        note:
           not usually instantiated directly by application code
        """

        self.model = parent_model
        self.node = patch_node
        self.patch_index = patch_index  # if not None and extracting from xml, patch_index must match xml
        self.triangle_count = 0
        self.node_count = 0
        self.triangles = None
        self.quad_triangles = None
        self.ni = None  # used to convert a triangle index back into a (j, i) pair when freshly built from mesh
        self.points = None
        self.crs_uuid = crs_uuid
        if patch_node is not None:
            xml_patch_index = rqet.find_tag_int(patch_node, 'PatchIndex')
            assert xml_patch_index is not None
            if self.patch_index is not None:
                assert self.patch_index == xml_patch_index, 'triangle patch index mismatch'
            else:
                self.patch_index = xml_patch_index
            self.triangle_count = rqet.find_tag_int(patch_node, 'Count')
            assert self.triangle_count is not None
            self.node_count = rqet.find_tag_int(patch_node, 'NodeCount')
            assert self.node_count is not None
            self.extract_crs_root_and_uuid()
            assert self.crs_uuid is not None

    def extract_crs_root_and_uuid(self):
        """Caches uuid for coordinate reference system, as stored in geometry xml sub-tree."""

        if self.crs_uuid is None:
            crs_root = rqet.find_nested_tags(self.node, ['Geometry', 'LocalCrs'])
            assert crs_root is not None, 'failed to find crs reference in triangulated patch xml'
            self.crs_uuid = bu.uuid_from_string(rqet.find_tag_text(crs_root, 'UUID'))
        else:
            crs_root = self.model.root_for_uuid(self.crs_uuid)
        return crs_root, self.crs_uuid

    def triangles_and_points(self):
        """Returns arrays representing the patch.

        Returns:
           Tuple (triangles, points):

           * triangles (int array of shape[:, 3]): integer indices into points array,
             being the nodes of the corners of the triangles
           * points (float array of shape[:, 3]): flat array of xyz points, indexed by triangles
        """
        if self.triangles is not None:
            return (self.triangles, self.points)
        assert self.triangle_count is not None and self.node_count is not None

        geometry_node = rqet.find_tag(self.node, 'Geometry')
        assert geometry_node is not None
        p_root = rqet.find_tag(geometry_node, 'Points')
        assert p_root is not None, 'Points xml node not found for triangle patch'
        assert rqet.node_type(p_root) == 'Point3dHdf5Array'
        h5_key_pair = self.model.h5_uuid_and_path_for_node(p_root, tag = 'Coordinates')
        if h5_key_pair is None:
            return (None, None)
        try:
            self.model.h5_array_element(h5_key_pair,
                                        cache_array = True,
                                        object = self,
                                        array_attribute = 'points',
                                        dtype = 'float')
        except Exception:
            log.error('hdf5 points failure for triangle patch ' + str(self.patch_index))
            raise
        triangles_node = rqet.find_tag(self.node, 'Triangles')
        h5_key_pair = self.model.h5_uuid_and_path_for_node(triangles_node)
        if h5_key_pair is None:
            log.warning('No Triangles found in xml for patch index: ' + str(self.patch_index))
            return (None, None)
        try:
            self.model.h5_array_element(h5_key_pair,
                                        cache_array = True,
                                        object = self,
                                        array_attribute = 'triangles',
                                        dtype = 'int')
        except Exception:
            log.error('hdf5 triangles failure for triangle patch ' + str(self.patch_index))
            raise
        return (self.triangles, self.points)

    def set_to_trimmed_patch(self, larger_patch, xyz_box = None, xy_polygon = None, internal = False):
        """Populate this (empty) patch with triangles and points that overlap with a trimming volume.

        arguments:
            larger_patch (TriangulatedPatch): the larger patch, a copy of which is to be trimmed
            xyz_box (numpy float array of shape (2, 3), optional): if present, a cuboid in xyz space
               against which to trim the patch
            xy_polygon (closed convex resqpy.lines.Polyline, optional): if present, an xy boundary
               against which to trim
            internal (bool, default False): if True, only those triangles where all three vertices
               are wtihin the trimming space are kept; if False, triangles with at least one vertex
               within the space are kept

        notes:
            at least one of xyz_box or xy_polygon must be present; if both are present, a triangle
            must be within both boundaries to survive the trimming;
            xyz_box and xy_polygon must be in the same crs as the larger patch
        """

        log.debug('trimming patch')
        assert xyz_box is not None or xy_polygon is not None
        if xyz_box is not None:
            log.debug(f'trim xyz_box: {xyz_box}')
        if xy_polygon is not None:
            log.debug(f'xy_polygon min xyz: {np.min(xy_polygon.coordinates, axis = 0)}')
            log.debug(f'xy_polygon max xyz: {np.max(xy_polygon.coordinates, axis = 0)}')
        large_t, large_p = larger_patch.triangles_and_points()
        log.debug(f'large surface min xyz: {np.min(large_p, axis = 0)}')
        log.debug(f'large surface max xyz: {np.max(large_p, axis = 0)}')
        # create bool per point indicating inclusion in box volume
        if xyz_box is None:
            points_in = np.ones(large_p.shape[:-1], dtype = bool)
        else:
            points_in = np.logical_and(np.all(large_p >= np.expand_dims(xyz_box[0], axis = 0), axis = -1),
                                       np.all(large_p <= np.expand_dims(xyz_box[1], axis = 0), axis = -1))
        # and check against xy polygon if present
        if xy_polygon is not None:
            points_in = np.logical_and(points_in, xy_polygon.points_are_inside_xy(large_p))
        # find where in large_t uses those points
        tp_in = points_in[large_t]
        t_in = np.all(tp_in, axis = -1) if internal else np.any(tp_in, axis = -1)
        # find unique points used by those triangles
        p_keep = np.unique(large_t[t_in])
        # note new point index for each old point that is being kept
        p_map = np.full(len(points_in), -1, dtype = int)
        p_map[p_keep] = np.arange(len(p_keep))
        # copy those unique points into a trimmed points array
        points_trimmed = large_p[p_keep]
        # copy selected triangles, replacing p indices with compressed indices
        t_trim = large_t[t_in]
        triangles_trimmed = p_map[t_trim]
        assert np.all(triangles_trimmed >= 0)
        assert np.all(triangles_trimmed < len(points_trimmed))
        self.crs_uuid = larger_patch.crs_uuid
        self.points = points_trimmed
        self.node_count = len(self.points)
        self.triangles = triangles_trimmed
        self.triangle_count = len(self.triangles)

    def set_to_horizontal_plane(self, depth, box_xyz, border = 0.0, quad_triangles = False):
        """Populate this (empty) patch with two triangles defining a flat, horizontal plane at a given depth.

        arguments:
           depth (float): z value to use in all points in the triangulated patch
           box_xyz (float[2, 3]): the min, max values of x, y (&z) giving the area to be covered (z ignored)
           border (float): an optional border width added around the x,y area defined by box_xyz
           quad_triangles (bool, default False): if True, 4 triangles are used instead of 2
        """

        # expand area by border
        box = box_xyz.copy()
        box[0, :2] -= border
        box[1, :2] += border
        self.node_count = 5 if quad_triangles else 4
        self.points = np.empty((self.node_count, 3))
        # set 4 points from corners of box
        self.points[0, :] = box[0, :]
        self.points[1, :] = box[1, :]
        self.points[2, 0], self.points[2, 1] = box[0, 0], box[1, 1]  # min x, max y
        self.points[3, 0], self.points[3, 1] = box[1, 0], box[0, 1]  # max x, min y
        if quad_triangles:
            self.points[4] = np.mean(box, axis = 0)
        # set depth for all points
        self.points[:, 2] = depth
        # create pair of triangles
        if quad_triangles:
            self.triangle_count = 4
            self.triangles = np.array([[0, 2, 4], [2, 1, 4], [1, 3, 4], [3, 0, 4]], dtype = int)
        else:
            self.triangle_count = 2
            self.triangles = np.array([[0, 1, 2], [0, 3, 1]], dtype = int)

    def set_to_triangle(self, corners):
        """Populate this (empty) patch with a single triangle."""

        assert corners.shape == (3, 3)
        self.node_count = 3
        self.points = corners.copy()
        self.triangle_count = 1
        self.triangles = np.array([[0, 1, 2]], dtype = int)

    def set_to_triangle_pair(self, corners):
        """Populate this (empty) patch with a pair of triangles."""

        self.set_from_triangles_and_points(np.array([[0, 1, 3], [0, 3, 2]], dtype = int), corners)

    def set_from_triangles_and_points(self, triangles, points):
        """Populate this (empty) patch from triangle node indices and points from elsewhere."""

        assert triangles.ndim == 2 and triangles.shape[-1] == 3
        assert points.ndim == 2 and points.shape[1] in [2, 3]
        if points.shape[1] == 2:
            p = np.zeros((points.shape[0], 3))
            p[:, :2] = points
            points = p
        self.node_count = points.shape[0]
        self.points = points.copy()
        self.triangle_count = triangles.shape[0]
        self.triangles = triangles.copy()

    def set_to_sail(self, n, centre, radius, azimuth, delta_theta):
        """Populate this (empty) patch with triangles for a big triangle wrapped on a sphere."""

        def sail_point(centre, radius, a, d):
            # m = vec.rotation_3d_matrix((d, 0.0, 90.0 - a))
            # m = vec.tilt_3d_matrix(a, d)
            v = np.array((0.0, radius, 0.0))
            m = vec.rotation_matrix_3d_axial(0, d)
            v = vec.rotate_vector(m, v)
            m = vec.rotation_matrix_3d_axial(2, 90.0 + a)
            v = vec.rotate_vector(m, v)
            return centre + v

        assert n >= 1
        self.node_count = (n + 1) * (n + 2) // 2
        self.points = np.empty((self.node_count, 3))
        self.triangle_count = n * n
        self.triangles = np.empty((self.triangle_count, 3), dtype = int)
        self.points[0] = sail_point(centre, radius, azimuth, 0.0).copy()
        p = 0
        t = 0
        for row in range(n):
            azimuth -= 0.5 * delta_theta
            dip = (row + 1) * delta_theta
            az = azimuth
            for pp in range(row + 2):
                p += 1
                self.points[p] = sail_point(centre, radius, az, dip).copy()
                az += delta_theta

            #           log.debug('p: ' + str(p) + '; dip: {0:3.1f}; az: {1:3.1f}'.format(dip, az))
            p1 = row * (row + 1) // 2
            p2 = (row + 1) * (row + 2) // 2
            self.triangles[t] = (p1, p2, p2 + 1)
            t += 1
            for tri in range(row):
                self.triangles[t] = (p1, p1 + 1, p2 + 1)
                t += 1
                p1 += 1
                p2 += 1
                self.triangles[t] = (p1, p2, p2 + 1)
                t += 1
        log.debug('sail point zero: ' + str(self.points[0]))
        log.debug('sail xyz box: {0:4.2f}:{1:4.2f}  {2:4.2f}:{3:4.2f}  {4:4.2f}:{5:4.2f}'.format(
            np.min(self.points[:, 0]), np.max(self.points[:, 0]), np.min(self.points[:, 1]), np.max(self.points[:, 1]),
            np.min(self.points[:, 2]), np.max(self.points[:, 2])))

    def set_from_irregular_mesh(self, mesh_xyz, quad_triangles = False):
        """Populate this (empty) patch from an untorn mesh array of shape (N, M, 3)."""

        mesh_shape = mesh_xyz.shape
        assert len(mesh_shape) == 3 and mesh_shape[0] > 1 and mesh_shape[1] > 1 and mesh_shape[2] == 3
        ni = mesh_shape[1]
        if quad_triangles:
            # generate centre points:
            quad_centres = np.empty(((mesh_shape[0] - 1) * (mesh_shape[1] - 1), 3))
            quad_centres[:, :] = 0.25 * (mesh_xyz[:-1, :-1, :] + mesh_xyz[:-1, 1:, :] + mesh_xyz[1:, :-1, :] +
                                         mesh_xyz[1:, 1:, :]).reshape((-1, 3))
            self.points = np.concatenate((mesh_xyz.copy().reshape((-1, 3)), quad_centres))
            mesh_size = mesh_xyz.size // 3
            self.node_count = self.points.size // 3
            self.triangle_count = 4 * (mesh_shape[0] - 1) * (mesh_shape[1] - 1)
            self.quad_triangles = True
            triangles = np.empty((mesh_shape[0] - 1, mesh_shape[1] - 1, 4, 3), dtype = int)  # flatten later
            nic = ni - 1
            for j in range(mesh_shape[0] - 1):
                for i in range(nic):
                    triangles[j, i, :, 0] = mesh_size + j * nic + i  # quad centre
                    triangles[j, i, 0, 1] = j * ni + i
                    triangles[j, i, 0, 2] = triangles[j, i, 1, 1] = j * ni + i + 1
                    triangles[j, i, 1, 2] = triangles[j, i, 2, 1] = (j + 1) * ni + i + 1
                    triangles[j, i, 2, 2] = triangles[j, i, 3, 1] = (j + 1) * ni + i
                    triangles[j, i, 3, 2] = j * ni + i
        else:
            self.points = mesh_xyz.copy().reshape((-1, 3))
            self.node_count = mesh_shape[0] * mesh_shape[1]
            self.triangle_count = 2 * (mesh_shape[0] - 1) * (mesh_shape[1] - 1)
            self.quad_triangles = False
            triangles = np.empty((mesh_shape[0] - 1, mesh_shape[1] - 1, 2, 3), dtype = int)  # flatten later
            for j in range(mesh_shape[0] - 1):
                for i in range(mesh_shape[1] - 1):
                    triangles[j, i, 0, 0] = j * ni + i
                    triangles[j, i, :, 1] = j * ni + i + 1
                    triangles[j, i, :, 2] = (j + 1) * ni + i
                    triangles[j, i, 1, 0] = (j + 1) * ni + i + 1
        self.ni = ni - 1
        self.triangles = triangles.reshape((-1, 3))

    def set_from_sparse_mesh(self, mesh_xyz):
        """Populate this (empty) patch from a mesh array of shape (N, M, 3), with some NaNs in z."""

        # todo: add quad_triangles argument to apply to 'squares' with no NaNs
        mesh_shape = mesh_xyz.shape
        assert len(mesh_shape) == 3 and mesh_shape[0] > 1 and mesh_shape[1] > 1 and mesh_shape[2] == 3
        self.quad_triangles = False

        indices = self.get_indices_from_sparse_meshxyz(mesh_xyz)

        triangles = np.zeros((2 * (mesh_shape[0] - 1) * (mesh_shape[1] - 1), 3), dtype = int)  # truncate later
        nt = 0
        for j in range(mesh_shape[0] - 1):
            for i in range(mesh_shape[1] - 1):
                nan_nodes = np.count_nonzero(np.isnan(mesh_xyz[j:j + 2, i:i + 2, 2]))
                if nan_nodes > 1:
                    continue
                if nan_nodes == 0:
                    triangles[nt, 0] = indices[j, i]
                    triangles[nt:nt + 2, 1] = indices[j, i + 1]
                    triangles[nt:nt + 2, 2] = indices[j + 1, i]
                    triangles[nt + 1, 0] = indices[j + 1, i + 1]
                    nt += 2
                elif indices[j, i] < 0:
                    triangles[nt, 0] = indices[j, i + 1]
                    triangles[nt, 1] = indices[j + 1, i]
                    triangles[nt, 2] = indices[j + 1, i + 1]
                    nt += 1
                elif indices[j, i + 1] < 0:
                    triangles[nt, 0] = indices[j, i]
                    triangles[nt, 1] = indices[j + 1, i]
                    triangles[nt, 2] = indices[j + 1, i + 1]
                    nt += 1
                elif indices[j + 1, i] < 0:
                    triangles[nt, 0] = indices[j, i + 1]
                    triangles[nt, 1] = indices[j, i]
                    triangles[nt, 2] = indices[j + 1, i + 1]
                    nt += 1
                elif indices[j + 1, i + 1] < 0:
                    triangles[nt, 0] = indices[j, i + 1]
                    triangles[nt, 1] = indices[j + 1, i]
                    triangles[nt, 2] = indices[j, i]
                    nt += 1
                else:
                    raise Exception('code failure in sparse mesh processing')
        self.ni = None
        self.triangles = triangles[:nt, :]
        self.triangle_count = nt

    def get_indices_from_sparse_meshxyz(self, mesh_xyz):
        """Update self.points and self.node_count with non-nan points in a given mesh_xyz array.

        Returns the indices of these non_nan points.
        """
        points = np.zeros(mesh_xyz.shape).reshape((-1, 3))
        indices = np.zeros(mesh_xyz.shape[:2], dtype = int) - 1

        non_nans = np.where(~np.isnan(mesh_xyz[:, :, 2]))
        for i in range(len(non_nans[0])):
            points[i] = mesh_xyz[non_nans[0][i], non_nans[1][i]]
            indices[non_nans[0][i], non_nans[1][i]] = i
        self.points = points[:len(non_nans[0]), :]
        self.node_count = len(non_nans[0])

        return indices

    def set_from_torn_mesh(self, mesh_xyz, quad_triangles = False):
        """Populate this (empty) patch from a torn mesh array of shape (nj, ni, 2, 2, 3)."""

        mesh_shape = mesh_xyz.shape
        assert len(mesh_shape) == 5 and mesh_shape[0] > 0 and mesh_shape[1] > 0 and mesh_shape[2:] == (2, 2, 3)
        nj = mesh_shape[0]
        ni = mesh_shape[1]
        if quad_triangles:
            # generate centre points:
            quad_centres = np.empty((nj, ni, 3))
            quad_centres[:, :, :] = 0.25 * np.sum(mesh_xyz, axis = (2, 3))
            self.points = np.concatenate((mesh_xyz.copy().reshape((-1, 3)), quad_centres.reshape((-1, 3))))
            mesh_size = mesh_xyz.size // 3
            self.node_count = 5 * nj * ni
            self.triangle_count = 4 * nj * ni
            self.quad_triangles = True
            triangles = np.empty((nj, ni, 4, 3), dtype = int)  # flatten later
            for j in range(nj):
                for i in range(ni):
                    base_p = 4 * (j * ni + i)
                    triangles[j, i, :, 0] = mesh_size + j * ni + i  # quad centre
                    triangles[j, i, 0, 1] = base_p
                    triangles[j, i, 0, 2] = triangles[j, i, 1, 1] = base_p + 1
                    triangles[j, i, 1, 2] = triangles[j, i, 2, 1] = base_p + 3
                    triangles[j, i, 2, 2] = triangles[j, i, 3, 1] = base_p + 2
                    triangles[j, i, 3, 2] = base_p
        else:
            self.points = mesh_xyz.copy().reshape((-1, 3))
            self.node_count = 4 * nj * ni
            self.triangle_count = 2 * nj * ni
            self.quad_triangles = False
            triangles = np.empty((nj, ni, 2, 3), dtype = int)  # flatten later
            for j in range(nj):
                for i in range(ni):
                    base_p = 4 * (j * ni + i)
                    triangles[j, i, 0, 0] = base_p
                    triangles[j, i, :, 1] = base_p + 1
                    triangles[j, i, :, 2] = base_p + 2
                    triangles[j, i, 1, 0] = base_p + 3
        self.ni = ni
        self.triangles = triangles.reshape((-1, 3))

    def column_from_triangle_index(self, triangle_index):
        """For patch freshly built from fully defined mesh, returns (j, i) for given triangle index.

        argument:
           triangle_index (int or numpy int array): the triangle index (or array of indices) for which column(s) are being
           sought

        returns:
           pair of ints or pair of numpy int arrays: the (j0, i0) indices of the column(s) which the triangle(s) is/are
           part of

        notes:
           this function will only work if the surface has been freshly constructed with data from a mesh without NaNs,
           otherwise (None, None) will be returned;
           if triangle_index is a numpy int array, a pair of similarly shaped numpy arrays is returned
        """

        if self.quad_triangles is None or self.ni is None:
            return (None, None)
        if isinstance(triangle_index, int):
            if triangle_index >= self.triangle_count:
                return (None, None)
        else:
            if np.any(triangle_index >= self.triangle_count):
                return (None, None)
        if self.quad_triangles:
            face = triangle_index // 4
        else:
            face = triangle_index // 2
        if isinstance(face, int):
            return divmod(face, self.ni)
        return np.divmod(face, self.ni)

    def set_to_cell_faces_from_corner_points(self, cp, quad_triangles = True):
        """Populates this (empty) patch to represent faces of a cell, from corner points of shape (2, 2, 2, 3)."""

        assert cp.shape == (2, 2, 2, 3)
        self.quad_triangles = quad_triangles
        if quad_triangles:
            triangles = self.get_triangles_for_cell_faces_quad_true(cp)
        else:
            triangles = self.get_triangles_for_cell_faces_quad_false(cp)
        self.triangles = triangles.reshape((-1, 3))

    def get_triangles_for_cell_faces_quad_false(self, cp):
        """Returns the triangles for corner points representing cell faces, where quad_triangles is False."""

        self.triangle_count = 12
        self.node_count = 8
        self.points = cp.copy().reshape((-1, 3))
        triangles = np.empty((3, 2, 2, 3), dtype = int)  # flatten later
        for axis in range(3):
            if axis == 0:
                ip1, ip2 = 2, 1
            elif axis == 1:
                ip1, ip2 = 4, 1
            else:
                ip1, ip2 = 4, 2
            for ip in range(2):
                ips = -2 * ip + 1  # +1 for -ve face; -1 for +ve face!
                base_p = 7 * ip
                triangles[axis, ip, :, 0] = base_p
                triangles[axis, ip, :, 1] = base_p + ips * (ip1 + ip2)
                triangles[axis, ip, 0, 2] = base_p + ips * (ip2)
                triangles[axis, ip, 1, 2] = base_p + ips * (ip1)

        return triangles

    def get_triangles_for_cell_faces_quad_true(self, cp):
        """Returns the triangles for corner points representing cell faces, where quad_triangles is True."""

        self.triangle_count = 24
        quad_centres = np.empty((3, 2, 3))
        quad_centres[0, 0, :] = 0.25 * np.sum(cp[0, :, :, :], axis = (0, 1))  # K-
        quad_centres[0, 1, :] = 0.25 * np.sum(cp[1, :, :, :], axis = (0, 1))  # K+
        quad_centres[1, 0, :] = 0.25 * np.sum(cp[:, 0, :, :], axis = (0, 1))  # J-
        quad_centres[1, 1, :] = 0.25 * np.sum(cp[:, 1, :, :], axis = (0, 1))  # J+
        quad_centres[2, 0, :] = 0.25 * np.sum(cp[:, :, 0, :], axis = (0, 1))  # I-
        quad_centres[2, 1, :] = 0.25 * np.sum(cp[:, :, 1, :], axis = (0, 1))  # I+
        self.node_count = 14
        self.points = np.concatenate((cp.copy().reshape((-1, 3)), quad_centres.reshape((-1, 3))))
        triangles = np.empty((3, 2, 4, 3), dtype = int)  # flatten later
        for axis in range(3):
            if axis == 0:
                ip1, ip2 = 2, 1
            elif axis == 1:
                ip1, ip2 = 4, 1
            else:
                ip1, ip2 = 4, 2
            for ip in range(2):
                centre_p = 8 + 2 * axis + ip
                ips = -2 * ip + 1  # +1 for -ve face; -1 for +ve face!
                base_p = 7 * ip
                triangles[axis, ip, :, 0] = centre_p  # quad centre
                triangles[axis, ip, 0, 1] = base_p
                triangles[axis, ip, 0, 2] = triangles[axis, ip, 1, 1] = base_p + ips * (ip2)
                triangles[axis, ip, 1, 2] = triangles[axis, ip, 2, 1] = base_p + ips * (ip1 + ip2)
                triangles[axis, ip, 2, 2] = triangles[axis, ip, 3, 1] = base_p + ips * (ip1)
                triangles[axis, ip, 3, 2] = base_p

        return triangles

    def face_from_triangle_index(self, triangle_index):
        """For patch freshly built for cell faces, returns (axis, polarity) for given triangle index."""

        if self.quad_triangles:
            assert self.node_count == 30 and self.triangle_count == 24
            assert 0 <= triangle_index < 24
            face = triangle_index // 4
        else:
            assert self.node_count == 24 and self.triangle_count == 12
            assert 0 <= triangle_index < 12
            face = triangle_index // 2
        axis, polarity = divmod(face, 2)
        return axis, polarity

    def vertical_rescale_points(self, ref_depth, scaling_factor):
        """Rescale points along vertical direction.
        
        Modifies the z values of points for this patch by stretching the distance
        from reference depth by scaling factor.
        """
        _, _ = self.triangles_and_points()  # ensure points are loaded
        z_values = self.points[:, 2].copy()
        self.points[:, 2] = ref_depth + scaling_factor * (z_values - ref_depth)



"""Multiprocessing wrapper functions for the Mesh class."""

import logging

log = logging.getLogger(__name__)

import uuid
from typing import Tuple, Union, List
from pathlib import Path
from uuid import UUID

import resqpy.model as rq
import resqpy.grid as grr
import resqpy.surface as rqs
from resqpy.multiprocessing import function_multiprocessing


def mesh_from_regular_grid_column_property_wrapper(
    index: int,
    parent_tmp_dir: str,
    grid_epc: str,
    grid_uuid: Union[UUID, str],
    prop_uuids: List[Union[UUID, str]],
) -> Tuple[int, bool, str, List[Union[UUID, str]]]:
    """Wrapper function of the Mesh from_regular_grid_column_property method.

    Used for multiprocessing to create a new model that is saved in a temporary epc file
    and returns the required values, which are used in the multiprocessing function to
    recombine all the objects into a single epc file.

    Args:
        index (int): the index of the function call from the multiprocessing function.
        parent_tmp_dir (str): the parent temporary directory path from the multiprocessing function.
        grid_epc (str): epc file path where the grid is saved.
        grid_uuid (UUID/str): UUID (universally unique identifier) of the regular grid object.
        prop_uuids (List[UUID/str]): a list of the property uuids used to create each Mesh
            and their relationship.

    Returns:
        Tuple containing:

            - index (int): the index passed to the function.
            - success (bool): True if all the Mesh objects could be created, False
              otherwise.
            - epc_file (str): the epc file path where the objects are stored.
            - uuid_list (List[UUID/str]): list of UUIDs of relevant objects.
    """
    uuid_list = []
    tmp_dir = Path(parent_tmp_dir) / f"{uuid.uuid4()}"
    tmp_dir.mkdir(parents = True, exist_ok = True)
    epc_file = str(tmp_dir / "wrapper.epc")
    model = rq.new_model(epc_file = epc_file, quiet = True)

    g_model = rq.Model(grid_epc, quiet = True)
    g_crs_uuid = g_model.uuid(obj_type = "LocalDepth3dCrs",
                              related_uuid = grid_uuid)  # todo: check this relationship exists

    if g_crs_uuid is not None:
        model.copy_uuid_from_other_model(g_model, g_crs_uuid)
        uuid_list.append(g_crs_uuid)
    model.copy_uuid_from_other_model(g_model, uuid = grid_uuid)

    for prop_uuid in prop_uuids:
        model.copy_uuid_from_other_model(g_model, uuid = prop_uuid)

    grid = grr.RegularGrid(parent_model = model, uuid = grid_uuid)

    success = True
    for prop_uuid in prop_uuids:
        mesh = rqs.Mesh.from_regular_grid_column_property(model, grid.uuid, prop_uuid)
        if mesh is None:
            success = False
            continue
        mesh.write_hdf5()
        mesh.create_xml()
        uuid_list.append(mesh.uuid)
        model.create_reciprocal_relationship_uuids(mesh.uuid, "sourceObject", prop_uuid, "detinationObject")

    model.store_epc(quiet = True)

    return index, success, epc_file, uuid_list


def mesh_from_regular_grid_column_property_batch(
    grid_epc: str,
    grid_uuid: Union[UUID, str],
    prop_uuids: List[Union[UUID, str]],
    recombined_epc: str,
    cluster,
    n_workers: int,
    require_success: bool = False,
) -> List[bool]:
    """Creates Mesh objects from a list of property uuids in parallel.

    Args:
        grid_epc (str): epc file path where the grid is saved.
        grid_uuid (UUID/str): UUID (universally unique identifier) of the grid object.
        prop_uuids (List[UUID/str]): a list of the column property uuids used to create each Mesh
            and their relationship.
        recombined_epc (Path/str): A pathlib Path or path string of
            where the combined epc will be saved.
        cluster (LocalCluster/JobQueueCluster): a LocalCluster is a Dask cluster on a
            local machine. If using a job queing system, a JobQueueCluster can be used
            such as an SGECluster, SLURMCluster, PBSCluster, LSFCluster etc.
        n_workers (int): the number of workers on the cluster.
        require_success (bool, default False): if True an exception is raised if any failures

    Returns:
        success_list (List[bool]): A boolean list of successful function calls.
    """
    n_uuids = len(prop_uuids)
    prop_uuids_list = [prop_uuids[i * n_uuids // n_workers:(i + 1) * n_uuids // n_workers] for i in range(n_workers)]

    kwargs_list = []
    for prop_uuids in prop_uuids_list:
        d = {
            "grid_epc": grid_epc,
            "grid_uuid": grid_uuid,
            "prop_uuids": prop_uuids,
        }
        kwargs_list.append(d)

    success_list = function_multiprocessing(mesh_from_regular_grid_column_property_wrapper,
                                            kwargs_list,
                                            recombined_epc,
                                            cluster,
                                            require_success = require_success)

    return success_list



"""uuid.py: Thin wrapper around python uuid (universally unique identifier) module."""

# NB: at present the code does not enforce multiprocessor safe generation of unique identifiers
# it calls uuid.uuid1() to generate new uuids, ie. using version 1 of the iso standard options

import uuid

test_mode = False
test_latest_int = 0

max_version_string_length = 10


def switch_on_test_mode(seed = 0):
    """Causes subsequent calls to new_uid() to produce integer sequence starting from successor to seed.

    arguments:
       seed (integer, default 0): The predecessor to the first uuid returned by subsequent calls to
          new_uuid()

    returns:
       None

    notes:
       call switch_off_test_mode() to reactivate normal behaviour;
       uuids generated whilst in test mode do not adhere to the iso standard;
       test mode is intended to allow replicatable behaviour for testing purposes
    """

    global test_mode
    global test_latest_int
    test_latest_int = seed
    test_mode = True


def switch_off_test_mode():
    """Subsequent calls to new_uid() will produce standard uuid values (default behaviour).

    note:
       this function will have no effect unless switch_on_test_mode() has previously been called
    """

    global test_mode
    test_mode = False


def new_uuid():
    """Returns a new uuid based on the time (to 100ns) & MAC address option of the iso standard.

    returns:
       uuid.UUID object

    notes:
       at present, the multi-processor safe functionality is not deployed, so multiple processors
       sharing the same MAC address could generate the same uuid simultaneously;
       an integer sequence is generated when in test mode
    """

    global test_latest_int
    if test_mode:
        test_latest_int += 1
        return uuid.UUID(bytes = test_latest_int.to_bytes(16, byte_order = 'big'))
    else:
        return uuid.uuid1()  # time to 100ns & MAC address


def string_from_uuid(uuid_obj):
    """Returns standard hexadecimal string for uuid; same as str(uuid_obj).

    arguments:
       uuid_obj (uuid.UUID object): the uuid which is required in hexadecimal string format

    returns:
       string (40 characters: 36 lowercase hexadecimals and 4 hyphens)
    """

    return str(uuid_obj)


def uuid_from_string(uuid_str):
    """Returns a uuid object for the given uuid string; hyphens are ignored.

    arguments:
       uuid_str (string): the hexadecimal representation of the 128 bit uuid integer;
          hyphens are ignored

    returns:
       uuid.UUID object

    notes:
       if a uuid.UUID object is passed by accident, it is returned;
       if the string starts with an underscore, the underscore is skipped (to cater for a fesapi quirk);
       any tail beyond the uuid string is ignored
    """

    if uuid_str is None:
        return None
    if isinstance(uuid_str, uuid.UUID):
        return uuid_str  # resilience to accidentally passing a uuid object
    try:
        if uuid_str[0] == '_':  # tolerate one of the fesapi quirks
            if len(uuid_str) < 37:
                return None
            return uuid.UUID(uuid_str[1:37])
        else:
            if len(uuid_str) < 36:
                return None
            return uuid.UUID(uuid_str[:36])
    except Exception:
        # could log or raise an error or warning?
        return None


def uuid_as_bytes(uuid_obj):
    """Returns the uuid as a 16 byte bytes sequence; same as uuid_obj.bytes.

    arguments:
       uuid_obj (uuid.UUID object): the uuid for which a bytes representation is required

    returns:
       bytes (16 bytes long)
    """

    if uuid_obj is None:
        return None
    if isinstance(uuid_obj, str):
        uuid_obj = uuid_from_string(uuid_obj)  # resilience to accidental string arg
    assert isinstance(uuid_obj, uuid.UUID)
    return uuid_obj.bytes


def uuid_as_int(uuid_obj):
    """Returns the uuid as a 128 bit int; same as uuid_obj.int.

    arguments:
       uuid_obj (uuid.UUID object): the uuid for which a bytes representation is required

    returns:
       bytes (16 bytes long)
    """

    if uuid_obj is None:
        return None
    if isinstance(uuid_obj, str):
        uuid_obj = uuid_from_string(uuid_obj)  # resilience to accidental string arg
    if not isinstance(uuid_obj, uuid.UUID):
        raise ValueError(f'non uuid object where uuid expected: {uuid_obj}; type: {type(uuid_obj)}')
    return uuid_obj.int


def matching_uuids(uuid_a, uuid_b):
    """Returns True if the 2 uuid objects are for the same id; False otherwise.

    arguments:
       uuid_a, uuid_b (uuid.UUID objects): the two uuids to be compared

    returns:
       boolean: True if the two uuids are the same; False otherwise

    note:
       this function is resilient to uuids being passed in hexadecimal string format
    """

    if isinstance(uuid_a, str):
        uuid_a = uuid_from_string(uuid_a)  # resilience to accidental string arg
    if isinstance(uuid_b, str):
        uuid_b = uuid_from_string(uuid_b)
    if uuid_a is None or uuid_b is None:
        return False
    return uuid_a.int == uuid_b.int


def version_string(uuid_obj):
    """Returns an integer string rendering of the time element of the uuid.

    arguments:
       uuid_obj (uuid.UUID): the uuid for which a string representation of the time component is required

    returns:
       string (of digits)

    notes:
       this function has nothing to do with the uuid.version attribute, it is used to populate the version
       field of a resqml citation block;
       the time component of the uuid is the number of 100ns periods that have elapsed since October 1582
       (when the Gregorian calendar was adopted), as a 60 bit integer
    """

    if isinstance(uuid_obj, str):
        uuid_obj = uuid_from_string(uuid_obj)  # resilience to accidental string arg
    v_str = str(uuid_obj.time)
    if len(v_str) > max_version_string_length:
        v_str = v_str[:max_version_string_length]
    return v_str


def is_uuid(uuid_obj):
    """Returns boolean indicating whether uuid_obj seems to be a uuid."""

    if isinstance(uuid_obj, uuid.UUID):
        return True
    if not uuid_obj or not isinstance(uuid_obj, str):
        return False
    if uuid_obj[0] == '_':
        return len(uuid_obj) == 37
    return len(uuid_obj) == 36
    # could also check for hyphens in correct places, and hexadecimals only



"""Multiprocessing wrapper functions for the well/BlockedWell class."""

import logging

log = logging.getLogger(__name__)

import uuid
from typing import Tuple, Union, List
from pathlib import Path
from uuid import UUID

import resqpy.model as rq
import resqpy.grid as grr
import resqpy.well as rqw
import resqpy.multiprocessing as rqmp


def blocked_well_from_trajectory_wrapper(
    index: int,
    parent_tmp_dir: str,
    grid_epc: str,
    grid_uuid: Union[UUID, str],
    trajectory_epc: str,
    trajectory_uuids: List[Union[UUID, str]],
) -> Tuple[int, bool, str, List[Union[UUID, str]]]:
    """Wrapper function of the BlockedWell initialisation from a Trajectory.

    Used for multiprocessing to create a new model that is saved in a temporary epc file
    and returns the required values, which are used in the multiprocessing function to
    recombine all the objects into a single epc file.

    Args:
        index (int): the index of the function call from the multiprocessing function.
        parent_tmp_dir (str): the parent temporary directory path from the multiprocessing function.
        grid_epc (str): epc file path where the grid is saved.
        grid_uuid (UUID/str): UUID (universally unique identifier) of the grid object.
        trajectory_epc (str): epc file path where the trajectories are saved.
        trajectory_uuids (List[UUID/str]): a list of the trajectory uuids used to create each
            Trajectory object.

    Returns:
        Tuple containing:

            - index (int): the index passed to the function.
            - success (bool): True if all the BlockedWell objects could be created, False
              otherwise.
            - epc_file (str): the epc file path where the objects are stored.
            - uuid_list (List[UUID/str]): list of UUIDs of relevant objects.
    """
    uuid_list = []
    tmp_dir = Path(parent_tmp_dir) / f"{uuid.uuid4()}"
    tmp_dir.mkdir(parents = True, exist_ok = True)
    epc_file = str(tmp_dir / "wrapper.epc")
    model = rq.new_model(epc_file = epc_file, quiet = True)

    trajectory_model = rq.Model(trajectory_epc, quiet = True)
    grid_model = rq.Model(grid_epc)
    model.copy_uuid_from_other_model(grid_model, uuid = grid_uuid)

    grid = grr.any_grid(grid_model, uuid = grid_uuid)

    success = True
    for trajectory_uuid in trajectory_uuids:
        model.copy_uuid_from_other_model(trajectory_model, uuid = trajectory_uuid)
        trajectory = rqw.Trajectory(model, trajectory_uuid)
        blocked_well = rqw.BlockedWell(model, grid = grid, trajectory = trajectory)
        if blocked_well is None or blocked_well.cell_count is None or blocked_well.node_count is None:
            success = False
            continue
        blocked_well.write_hdf5()
        blocked_well.create_xml()
        uuid_list.append(blocked_well.uuid)

    model.store_epc(quiet = True)

    return index, success, epc_file, uuid_list


def blocked_well_from_trajectory_batch(
    grid_epc: str,
    grid_uuid: Union[UUID, str],
    trajectory_epc: str,
    trajectory_uuids: List[Union[UUID, str]],
    recombined_epc: str,
    cluster,
    n_workers: int,
    require_success: bool = False,
) -> List[bool]:
    """Creates BlockedWell objects from a common grid and a list of trajectories uuids in parallel.

    Args:
        grid_epc (str): epc file path where the grid is saved.
        grid_uuid (UUID/str): UUID (universally unique identifier) of the grid object.
        trajectory_epc (str): epc file path where the trajectories are saved.
        trajectory_uuids (List[UUID/str]): a list of the trajectory uuids used to create each
            Trajectory object.
        recombined_epc (Path/str): A pathlib Path or path string of
            where the combined epc will be saved.
        cluster (LocalCluster/JobQueueCluster): a LocalCluster is a Dask cluster on a
            local machine. If using a job queing system, a JobQueueCluster can be used
            such as an SGECluster, SLURMCluster, PBSCluster, LSFCluster etc.
        n_workers (int): the number of workers on the cluster.
        require_success (bool, default False): if True an exception is raised if any failures

    Returns:
        success_list (List[bool]): A boolean list of successful function calls.

    Note:
        the returned success list contains one value per batch, set True if all blocked wells
        were successfully created in the batch, False if one or more failed in the batch
    """
    n_uuids = len(trajectory_uuids)
    trajectory_uuids_list = [
        trajectory_uuids[i * n_uuids // n_workers:(i + 1) * n_uuids // n_workers] for i in range(n_workers)
    ]

    kwargs_list = []
    for trajectory_uuids in trajectory_uuids_list:
        d = {
            "grid_epc": grid_epc,
            "grid_uuid": grid_uuid,
            "trajectory_epc": trajectory_epc,
            "trajectory_uuids": trajectory_uuids,
        }
        kwargs_list.append(d)

    success_list = rqmp.function_multiprocessing(blocked_well_from_trajectory_wrapper,
                                                 kwargs_list,
                                                 recombined_epc,
                                                 cluster,
                                                 require_success = require_success)

    return success_list



"""Miscellaneous functions relating to grids."""

# Nexus is a registered trademark of the Halliburton Company

import logging

log = logging.getLogger(__name__)

import math as maths
import numpy as np
import random

import resqpy.olio.factors as factors
import resqpy.olio.vector_utilities as vec


def infill_block_geometry(extent,
                          depth,
                          thickness,
                          x,
                          y,
                          k_increase_direction = 'down',
                          depth_zero_tolerance = 0.01,
                          x_y_zero_tolerance = 0.01,
                          vertical_cell_overlap_tolerance = 0.01,
                          snap_to_top_and_base = True,
                          nudge = True):
    """Scans logically vertical columns of cells setting depth (& thickness) of inactive cells.

    args:
       extent (numpy integer vector of shape (3,)): corresponds to nk, nj and ni
       depth (3D numpy float array): size matches extent.
        note: Depth values are assumed more positive with increasing depth. Zero values indicate inactive cells
       thickness (3D numpy float array): size matches extent
       x (3D numpy float array): size matches extent
       y (3D numpy float array): size matches extent
       k_increase_direction (string, default 'down'): direction of increasing values. Either 'up' or 'down'
       depth_zero_tolerance (float, optional, default 0.01): maximum value for which the depth is considered zero
       vertical_cell_overlap_tolerance (float, optional, default 0.01): maximum acceptable overlap of cells on input
       snap_to_top_and_base (boolean, optional, default True): when True, causes cells above topmost active and below
        deepest active to be populated with pinched out cells at the top and bottom faces respectively
        nudge (boolean, optional, default True): when True causes the depth of cells with greater k to be moved to
         clean up overlap over pinchouts

    """

    k_dir_sign = __get_k_dir_sign(k_increase_direction = k_increase_direction)

    for j in range(extent[1]):
        for i in range(extent[2]):
            k_top = 0  # NB: 'top' & 'bottom' are misleading if k_increase_direction == 'up'
            k_top, x, y, depth, thickness, whole_column_inactive = __clean_up_tiny_values(
                k = k_top,
                x = x,
                y = y,
                i = i,
                j = j,
                extent = extent,
                depth = depth,
                thickness = thickness,
                depth_zero_tolerance = depth_zero_tolerance,
                x_y_zero_tolerance = x_y_zero_tolerance)
            if whole_column_inactive:
                continue
            x, y, depth = __snap_to_top_and_base(snap_to_top_and_base = snap_to_top_and_base,
                                                 x = x,
                                                 y = y,
                                                 i = i,
                                                 j = j,
                                                 depth = depth,
                                                 thickness = thickness,
                                                 k_top = k_top,
                                                 k_dir_sign = k_dir_sign)
            while True:
                while k_top < extent[0] and abs(depth[k_top, j, i]) > depth_zero_tolerance:  # skip active layers
                    k_top += 1
                k_base = k_top + 1
                k_base, x, y, depth, thickness, _ = __clean_up_tiny_values(k = k_base,
                                                                           x = x,
                                                                           y = y,
                                                                           i = i,
                                                                           j = j,
                                                                           extent = extent,
                                                                           depth = depth,
                                                                           thickness = thickness,
                                                                           depth_zero_tolerance = depth_zero_tolerance,
                                                                           x_y_zero_tolerance = x_y_zero_tolerance)
                if k_base >= extent[0]:  # no deeper active cells found
                    x, y, depth = __snap_to_top_and_base(snap_to_top_and_base = snap_to_top_and_base,
                                                         x = x,
                                                         y = y,
                                                         i = i,
                                                         j = j,
                                                         depth = depth,
                                                         thickness = thickness,
                                                         k_top = k_top,
                                                         k_dir_sign = k_dir_sign,
                                                         k_base_greater_or_equal_to_extent = True)
                    break
                void_cell_count = k_base - k_top
                assert (void_cell_count > 0)
                void_top_depth = depth[k_top - 1, j, i] + (thickness[k_top - 1, j, i] / 2.0) * k_dir_sign
                void_bottom_depth = depth[k_base, j, i] - (thickness[k_base, j, i] / 2.0) * k_dir_sign
                void_top_x = x[k_top - 1, j, i]
                void_top_y = y[k_top - 1, j, i]
                void_interval = k_dir_sign * (void_bottom_depth - void_top_depth)
                void_x_interval = x[k_base, j, i] - void_top_x
                void_y_interval = y[k_base, j, i] - void_top_y
                infill_cell_thickness = void_interval / void_cell_count
                if void_interval < 0.0:  # overlapping cells
                    if -void_interval < vertical_cell_overlap_tolerance:
                        depth, infill_cell_thickness, void_interval = __nudge_overlapping_cells_if_requested(
                            nudge = nudge,
                            i = i,
                            j = j,
                            k_base = k_base,
                            extent = extent,
                            void_interval = void_interval,
                            void_bottom_depth = void_bottom_depth,
                            depth_zero_tolerance = depth_zero_tolerance,
                            k_dir_sign = k_dir_sign)
                    else:
                        log.warning('Cells [%1d, %1d, %1d] and [%1d, %1d, %1d] overlap ...', i + 1, j + 1, k_top, i + 1,
                                    j + 1, k_base + 1)
                        log.warning('Check k_increase_direction and tolerances')
                        log.warning('Skipping rest of i,j column')  # todo: could abort here
                        break
                assert (infill_cell_thickness >= 0.0)
                for void_k in range(void_cell_count):
                    depth[k_top + void_k, j, i] = void_top_depth + (0.5 + void_k) * infill_cell_thickness * k_dir_sign
                    thickness[k_top + void_k, j, i] = infill_cell_thickness
                    x[k_top + void_k, j, i] = void_top_x + (0.5 + void_k) * void_x_interval / void_cell_count
                    y[k_top + void_k, j, i] = void_top_y + (0.5 + void_k) * void_y_interval / void_cell_count
                k_top = k_base


def __get_k_dir_sign(k_increase_direction):
    """ Set whether depth increases with increasingly positive or negative values."""

    if k_increase_direction == 'down':
        k_dir_sign = 1.0
    elif k_increase_direction == 'up':
        k_dir_sign = -1.0
    else:
        assert (False)
    return k_dir_sign


def __clean_up_tiny_values(k, x, y, i, j, extent, depth, thickness, depth_zero_tolerance, x_y_zero_tolerance):
    """ Set x, y, depth and thickness values to zero if values are below tolerances."""

    whole_column_inactive = False
    while k < extent[0] and abs(depth[k, j, i]) <= depth_zero_tolerance:
        depth[k, j, i] = 0.0  # clean up tiny values
        thickness[k, j, i] = 0.0
        if abs(x[k, j, i]) <= x_y_zero_tolerance:
            x[k, j, i] = 0.0
        if abs(y[k, j, i]) <= x_y_zero_tolerance:
            y[k, j, i] = 0.0
        k += 1  # skip topmost inactive batch
    if k >= extent[0]:
        whole_column_inactive = True
    return k, x, y, depth, thickness, whole_column_inactive


def __snap_to_top_and_base(snap_to_top_and_base,
                           x,
                           y,
                           i,
                           j,
                           depth,
                           thickness,
                           k_top,
                           k_dir_sign,
                           k_base_greater_or_equal_to_extent = False):
    """ Cells above topmost active and below deepest active will be populated with pinched out cells at the top and

    bottom faces respectively.
    """
    if k_base_greater_or_equal_to_extent:
        k_top = k_top - 1
        if snap_to_top_and_base:
            snap_depth = depth[k_top, j, i] - k_dir_sign * thickness[k_top, j, i] / 2.0
            snap_x = x[k_top, j, i]
            snap_y = y[k_top, j, i]
            for k_snap in range(k_top):
                depth[k_snap, j, i] = snap_depth
                x[k_snap, j, i] = snap_x
                y[k_snap, j, i] = snap_y
    return x, y, depth


def __nudge_overlapping_cells_if_requested(nudge, i, j, k_base, extent, depth, void_interval, void_bottom_depth,
                                           depth_zero_tolerance, k_dir_sign):
    """ Clean up overlap over pinchouts by moving the depths of cells with greater k."""

    if nudge:
        nudge_count = 0  # debug
        for k_nudge in range(extent[0] - k_base):
            if depth[k_base + k_nudge, j, i] > depth_zero_tolerance:
                depth[k_base + k_nudge, j, i] += -void_interval * k_dir_sign
                nudge_count += 1  # debug
        log.debug('%1d cells nudged in [ i j ] column [%1d, %1d]', nudge_count, i + 1, j + 1)
        void_bottom_depth += -void_interval
    void_interval = 0.0
    infill_cell_thickness = 0.0
    return depth, infill_cell_thickness, void_interval


def resequence_nexus_corp(corner_points, eight_mode = False, undo = False):
    """Reorders corner point data in situ, to handle bizarre nexus orderings."""

    # undo False for corp to internal; undo True for internal to corp; only relevant in eight_mode
    assert (corner_points.ndim == 7)
    extent = np.array(corner_points.shape, dtype = 'int')
    if eight_mode:
        for k in range(extent[0]):
            for j in range(extent[1]):
                for i in range(extent[2]):
                    if undo:
                        xyz = np.zeros((3, 8))
                        c = 0
                        for kp in range(2):
                            for jp in range(2):
                                for ip in range(2):
                                    xyz[:, c] = corner_points[k, j, i, kp, jp, ip, :]
                                    c += 1
                        corner_points[k, j, i] = xyz.reshape((2, 2, 2, 3))
                    else:
                        xyz = corner_points[k, j, i].reshape((3, 8)).copy()
                        c = 0
                        for kp in range(2):
                            for jp in range(2):
                                for ip in range(2):
                                    corner_points[k, j, i, kp, jp, ip, :] = xyz[:, c]
                                    c += 1
    else:  # reversible, so not dependent on undo argument
        jp_slice = corner_points[:, :, :, :, 1, 0, :].copy()
        corner_points[:, :, :, :, 1, 0, :] = corner_points[:, :, :, :, 1, 1, :]
        corner_points[:, :, :, :, 1, 1, :] = jp_slice


def random_cell(corner_points, border = 0.25, max_tries = 20, tolerance = 0.003):
    """Returns a random cell's (k,j,i) tuple for a cell with non-zero lengths on all 3 primary edges."""

    assert (corner_points.ndim == 7)
    assert (border >= 0.0 and border < 0.5)
    assert (max_tries > 0)

    extent = np.array(corner_points.shape, dtype = 'int')
    kji_extent = extent[:3]
    kji_border = np.zeros(3, dtype = 'int')
    kji_upper = np.zeros(3, dtype = 'int')
    for axis in range(3):
        kji_border[axis] = int(float(kji_extent[axis]) * border)
        kji_upper[axis] = kji_extent[axis] - kji_border[axis] - 1
        if kji_upper[axis] < kji_border[axis]:
            kji_upper[axis] = kji_border[axis]

    kji_cell = np.empty(3, dtype = 'int')
    attempt = 0
    while attempt < max_tries:
        attempt += 1
        for axis in range(3):
            if kji_extent[axis] == 1:
                kji_cell[axis] = 0
            else:
                kji_cell[axis] = random.randint(kji_border[axis], kji_upper[axis])
        cell_cp = corner_points[tuple(kji_cell)]
        assert (cell_cp.shape == (2, 2, 2, 3))
        if vec.manhatten_distance(cell_cp[0, 0, 0], cell_cp[1, 0, 0]) < tolerance:
            continue
        if vec.manhatten_distance(cell_cp[0, 0, 0], cell_cp[0, 1, 0]) < tolerance:
            continue
        if vec.manhatten_distance(cell_cp[0, 0, 0], cell_cp[0, 0, 1]) < tolerance:
            continue
        return tuple(kji_cell)

    log.warning('failed to find random voluminous cell')
    return None


def determine_corp_ijk_handedness(corner_points, xyz_is_left_handed = True):
    """Determine true ijk handedness from corner point data in pagoda style 7D array; returns 'right' or 'left'."""

    assert (corner_points.ndim == 7)
    cell_kji = random_cell(corner_points)
    assert (cell_kji is not None)
    log.debug('using cell ijk0 [{}, {}, {}] to determine ijk handedness'.format(cell_kji[2], cell_kji[1], cell_kji[0]))
    cell_cp = corner_points[cell_kji]
    origin = cell_cp[0, 0, 0]
    det = vec.determinant(cell_cp[0, 0, 1] - origin, cell_cp[0, 1, 0] - origin,
                          cell_cp[1, 0, 0] - origin)  # NB. IJK ordering
    if det == 0.0:
        log.warning('indeterminate handedness in cell ijk0 [{}, {}, {}]'.format(cell_kji[2], cell_kji[1], cell_kji[0]))
        return None
    if det > 0.0:
        ijk_is_left_handed = xyz_is_left_handed
    else:
        ijk_is_left_handed = not xyz_is_left_handed
    if ijk_is_left_handed:
        return 'left'
    return 'right'


def determine_corp_extent(corner_points, tolerance = 0.003):
    """Returns extent of grid derived from 7D corner points with all cells temporarily in I."""

    def neighbours(corner_points, sextuple_cell_a_p1, sextuple_cell_a_p2, sextuple_cell_b_p1, sextuple_cell_b_p2,
                   tolerance):
        # allows for reversal of points (or not) in neighbouring cell
        if ((vec.manhatten_distance(corner_points[sextuple_cell_a_p1], corner_points[sextuple_cell_b_p1]) <= tolerance)
                and (vec.manhatten_distance(corner_points[sextuple_cell_a_p2], corner_points[sextuple_cell_b_p2]) <=
                     tolerance)):
            return True
        if ((vec.manhatten_distance(corner_points[sextuple_cell_a_p1], corner_points[sextuple_cell_b_p2]) <= tolerance)
                and (vec.manhatten_distance(corner_points[sextuple_cell_a_p2], corner_points[sextuple_cell_b_p1]) <=
                     tolerance)):
            return True
        return False

    assert (corner_points.ndim == 7 and corner_points.shape[:2] == (1, 1))

    confirmation = 3  # number of identical results needed for each of NI and NJ
    max_failures = 100  # maximum number of failed random cells for each of NI and NJ
    min_cell_length = 10.0 * tolerance
    border = 0.0 if corner_points.shape[2] < 1000 else 0.25

    cell_count = corner_points.shape[2]
    prime_factorization = factors.factorize(cell_count)
    log.debug('cell count is ' + str(cell_count) + '; prime factorization: ' + str(prime_factorization))
    possible_extents = factors.all_factors_from_primes(prime_factorization)
    log.debug('possible extents are: ' + str(possible_extents))

    ni = None
    redundancy = confirmation
    remaining_attempts = max_failures
    while redundancy:
        kji_cell = random_cell(corner_points, border = border, tolerance = min_cell_length)
        found = False
        if kji_cell is not None:
            for e in possible_extents:
                candidate = kji_cell[2] + e
                if candidate >= cell_count:
                    continue
                if neighbours(corner_points, (0, 0, kji_cell[2], 0, 1, 0), (0, 0, kji_cell[2], 0, 1, 1),
                              (0, 0, candidate, 0, 0, 0), (0, 0, candidate, 0, 0, 1), tolerance):
                    if ni is not None and ni != e:
                        log.error('inconsistent NI values of {} and {} determined from corner points'.format(ni, e))
                        return None
                    found = True
                    ni = e
                    redundancy -= 1
                    break
        if not found:
            remaining_attempts -= 1
            if remaining_attempts <= 0:
                log.error('failed to determine NI from corner points (out of tries)')  # could assume NJ = 1 here
                return None

    log.info('NI determined from corner points to be ' + str(ni))

    if ni > 1:
        ni_prime_factors = factors.factorize(ni)
        factors.remove_subset(prime_factorization, ni_prime_factors)
        log.debug('remaining prime factors after accounting for NI are: ' + str(prime_factorization))
        possible_extents = factors.all_factors_from_primes(prime_factorization)
        log.debug('possible extents for NJ & NK are: ' + str(possible_extents))

    nj = None
    redundancy = confirmation
    remaining_attempts = max_failures
    while redundancy:
        kji_cell = random_cell(corner_points)
        found = False
        for e in possible_extents:
            candidate = kji_cell[2] + (e * ni)
            if candidate >= cell_count:
                continue
            if vec.manhatten_distance(corner_points[0, 0, kji_cell[2], 1, 0, 0], corner_points[0, 0, candidate, 0, 0,
                                                                                               0]) <= tolerance:
                if nj is not None and nj != e:
                    log.error('inconsistent NJ values of {} and {} determined from corner points'.format(nj, e))
                    return None
                found = True
                nj = e
                redundancy -= 1
                break
        if not found:
            remaining_attempts -= 1
            if remaining_attempts <= 0:
                log.error(
                    'failed to determine NJ from corner points (out of tries)')  # could assume or check if NK = 1 here
                return None

    log.info('NJ determined from corner points to be ' + str(nj))

    nk, remainder = divmod(cell_count, ni * nj)
    assert (remainder == 0)
    log.info('NK determined from corner points to be ' + str(nk))
    assert (nk in possible_extents)

    return [nk, nj, ni]


def translate_corp(corner_points, x_shift = None, y_shift = None, min_xy = None, preserve_digits = None):
    """Adjusts x and y values of corner points by a constant offset."""

    assert (corner_points.ndim == 7)
    if min_xy is None:
        minimum_xy = 0.0
    else:
        minimum_xy = min_xy
    if x_shift is None:
        x_sub = np.min(corner_points[:, :, :, :, :, :, 0]) - minimum_xy
    else:
        x_sub = -x_shift
    if y_shift is None:
        y_sub = np.min(corner_points[:, :, :, :, :, :, 1]) - minimum_xy
    else:
        y_sub = -y_shift
    if preserve_digits is not None:
        divisor = maths.pow(10.0, preserve_digits)
        x_sub = divisor * maths.floor(x_sub / divisor)
        y_sub = divisor * maths.floor(y_sub / divisor)

    log.info('translating corner points by %3.1f in x and %3.1f in y', -x_sub, -y_sub)
    corner_points[:, :, :, :, :, :, 0] -= x_sub
    corner_points[:, :, :, :, :, :, 1] -= y_sub


def triangles_for_cell_faces(cp):
    """Returns numpy array of shape (3, 2, 4, 3, 3) with axes being kji, -+, triangle within face, triangle corner, xyz.

    args:
       cp (numpy float array of shape (2, 2, 2, 3)): single cell corner point array in pagoda protocol

    returns:
       numpy float array of shape (3, 2, 4, 3, 3) holding triangle corner coordinates for cell faces represented with
       quad triangles

    note:
       resqpy.surface also contains methods for working with cell faces as triangulated sets
    """

    tri = np.empty((3, 2, 4, 3, 3))

    # create face centre points and assign as one vertex in each of 4 trangles for face
    tri[0, :, :, 0] = np.mean(cp, axis = (1, 2)).reshape((2, 1, 3)).repeat(4, axis = 1).reshape(
        (2, 4, 3))  # k face centres
    tri[1, :, :, 0] = np.mean(cp, axis = (0, 2)).reshape((2, 1, 3)).repeat(4, axis = 1).reshape(
        (2, 4, 3))  # j face centres
    tri[2, :, :, 0] = np.mean(cp, axis = (0, 1)).reshape((2, 1, 3)).repeat(4, axis = 1).reshape(
        (2, 4, 3))  # i face centres

    # k faces
    tri[0, :, 0, 1] = cp[:, 0, 0]
    tri[0, :, 0, 2] = cp[:, 0, 1]
    tri[0, :, 1, 1] = cp[:, 0, 1]
    tri[0, :, 1, 2] = cp[:, 1, 1]
    tri[0, :, 2, 1] = cp[:, 1, 1]
    tri[0, :, 2, 2] = cp[:, 1, 0]
    tri[0, :, 3, 1] = cp[:, 1, 0]
    tri[0, :, 3, 2] = cp[:, 0, 0]

    # j faces
    tri[1, :, 0, 1] = cp[0, :, 0]
    tri[1, :, 0, 2] = cp[0, :, 1]
    tri[1, :, 1, 1] = cp[0, :, 1]
    tri[1, :, 1, 2] = cp[1, :, 1]
    tri[1, :, 2, 1] = cp[1, :, 1]
    tri[1, :, 2, 2] = cp[1, :, 0]
    tri[1, :, 3, 1] = cp[1, :, 0]
    tri[1, :, 3, 2] = cp[0, :, 0]

    # i faces
    tri[2, :, 0, 1] = cp[0, 0, :]
    tri[2, :, 0, 2] = cp[0, 1, :]
    tri[2, :, 1, 1] = cp[0, 1, :]
    tri[2, :, 1, 2] = cp[1, 1, :]
    tri[2, :, 2, 1] = cp[1, 1, :]
    tri[2, :, 2, 2] = cp[1, 0, :]
    tri[2, :, 3, 1] = cp[1, 0, :]
    tri[2, :, 3, 2] = cp[0, 0, :]

    return tri


def actual_pillar_shape(pillar_points, tolerance = 0.001):
    """Returns 'curved', 'straight' or 'vertical' for shape of pillar points.

    Args:
        pillar_points: fully defined points array of shape (nk + k_gaps + 1,..., 3).
    """

    assert pillar_points.ndim >= 3 and pillar_points.shape[-1] == 3

    pp = pillar_points.reshape((pillar_points.shape[0], -1, 3))

    from_top = pp - pp[0]
    xy_drift = np.abs(from_top[:, :, 0]) + np.abs(
        from_top[:, :, 1])  # use Manhattan distance as cheap proxy for true distance
    if np.max(xy_drift) <= tolerance:
        return 'vertical'
    if np.max(xy_drift[-1]) <= tolerance:
        return 'curved'  # top & bottom are vertically aligned, so pillar must be curved

    # where z variation is tiny (null pillar), don't interpolate, just treat these pillars as straight
    # elsewhere find drift from vector defined by from_top[-1]
    null_pillar_mask = (abs(from_top[-1, :, 2]) <= tolerance)
    from_top[-1, :, 2] = np.where(null_pillar_mask, tolerance, from_top[-1, :, 2])  # avoid divide by zero issues
    z_fraction = from_top[:, :, 2] / from_top[-1, :, 2]
    xy_drift = from_top[:, :, :2] - z_fraction.reshape((pp.shape[0], pp.shape[1], 1)) * from_top[-1, :, :2].reshape(
        (1, pp.shape[1], 2))
    straight = (np.max(np.sum(np.abs(xy_drift), axis = -1), axis = 0) <= tolerance)
    masked_straight = np.where(null_pillar_mask, True, straight)
    if np.all(masked_straight):
        return 'straight'
    return 'curved'


def columns_to_nearest_split_face(grid):
    """Return an int array of shape (NJ, NI) being number of cells to nearest split edge.

    Uses Manhattan distance.
    """
    if not grid.has_split_coordinate_lines:
        return None

    j_col_faces_split, i_col_faces_split = grid.split_column_faces()
    abutting = np.zeros((grid.nj, grid.ni), dtype = bool)
    abutting[:-1, :] = j_col_faces_split
    abutting[1:, :] = np.logical_or(abutting[1:, :], j_col_faces_split)
    abutting[:, :-1] = np.logical_or(abutting[:, :-1], i_col_faces_split)
    abutting[:, 1:] = np.logical_or(abutting[:, 1:], i_col_faces_split)
    framed = np.full((grid.nj + 2, grid.ni + 2), grid.nj + grid.ni, dtype = int)
    framed[1:-1, 1:-1] = np.where(abutting, 0, grid.nj + grid.ni)

    while True:
        plus_one = framed + 1
        updated = np.minimum(framed[1:-1, 1:-1], plus_one[:-2, 1:-1])
        updated[:] = np.minimum(updated, plus_one[2:, 1:-1])
        updated[:] = np.minimum(updated, plus_one[1:-1, :-2])
        updated[:] = np.minimum(updated, plus_one[1:-1, 2:])
        if np.all(updated == framed[1:-1, 1:-1]):
            break
        framed[1:-1, 1:-1] = updated

    return framed[1:-1, 1:-1]


def left_right_foursome(full_pillar_list, p_index):
    """Returns (2, 2) bool numpy array indicating which columns around a primary pillar are to the right of a line."""

    assert 0 < p_index < len(full_pillar_list) - 1
    here = np.array(full_pillar_list[p_index], dtype = int)
    previous = np.array(full_pillar_list[p_index - 1], dtype = int)
    next = np.array(full_pillar_list[p_index + 1], dtype = int)
    entry = tuple(here - previous)
    exit = tuple(next - here)
    # yapf: disable
    entry_tuples_list = [(0, 1), (0, -1), (1, 0), (-1, 0)]
    exit_tuples_list = [[(-1, 0), (0, 1),  (1, 0)],
                        [(-1, 0), (0, -1), (1, 0)],
                        [(0, -1), (1, 0),  (0, 1)],
                        [(0, -1), (-1, 0), (0, 1)]]
    result_arrays_list = [[np.array([[False, True],  [True,  True]],  dtype = bool),
                           np.array([[False, False], [True,  True]],  dtype = bool),
                           np.array([[False, False], [True,  False]], dtype = bool)],
                          [np.array([[False, True],  [False, False]], dtype = bool),
                           np.array([[True,  True],  [False, False]], dtype = bool),
                           np.array([[True,  True],  [True,  False]], dtype = bool)],
                          [np.array([[True,  False], [False, False]], dtype = bool),
                           np.array([[True,  False], [True,  False]], dtype = bool),
                           np.array([[True,  False], [True,  True]],  dtype = bool)],
                          [np.array([[True,  True],  [False, True]],  dtype = bool),
                           np.array([[False, True],  [False, True]],  dtype = bool),
                           np.array([[False, False], [False, True]],  dtype = bool)]]
    # yapf: enable
    try:
        list_index = entry_tuples_list.index(entry)
    except ValueError:
        log.debug(f'entry pair: {entry}')
        raise Exception('code failure whilst taking entry sides from dubious full pillar list')
    try:
        result_array_index = exit_tuples_list[list_index].index(exit)
        result_array = result_arrays_list[list_index][result_array_index]
        return result_array
    except ValueError:
        raise Exception('code failure whilst taking exit sides from dubious full pillar list')



"""_context.py: ModelContext class."""

import logging

log = logging.getLogger(__name__)

import os
from typing import Optional

import resqpy.model
from resqpy.model import Model, new_model


class ModelContext:
    """Context manager for easy opening and closing of resqpy models.

    When a model is opened this way, any open file handles are safely closed
    when the "with" clause exits. Optionally, the epc can be written back to
    disk upon exit.

    Example::

        with ModelContext("my_model.epc", mode="rw") as model:
            print(model.uuids())

    Note:
        The "write_hdf5" and "create_xml" methods of individual resqpy objects
        still need to be invoked as usual.
    """

    def __init__(self, epc_file, mode = "r") -> None:
        """Open a resqml file, safely closing file handles upon exit.

        The modes operate as follows:

        - In "read" mode, an existing epc file is opened. Any changes are not
        saved to disk automatically, but can still be saved by calling
        `model.store_epc()`.
        - In "read/write" mode, changes are written to disk when the context exists.
        - In "create" mode, a new model is created and saved upon exit. Any existing
            model will be deleted.

        Args:
            epc_file (str): path to existing resqml file
            mode (str): one of "read", "read/write", "create", or shorthands
            "r", "rw", "c".
        """

        # Validate mode
        modes_mapping = {"r": "read", "rw": "read/write", "c": "create"}
        mode = modes_mapping.get(mode, mode)
        if mode not in modes_mapping.values():
            raise ValueError(f"Unexpected mode '{mode}'")

        self.epc_file = epc_file
        self.mode = mode
        self._model: Optional[Model] = None

    def __enter__(self) -> Model:
        """Enter the runtime context, return a model."""

        if self.mode in ["read", "read/write"]:
            if not os.path.exists(self.epc_file):
                raise FileNotFoundError(self.epc_file)
            self._model = Model(epc_file = str(self.epc_file))

        else:
            assert self.mode == "create"
            for file in [self.epc_file, self.epc_file[:-4] + '.h5']:
                if os.path.exists(file):
                    os.remove(file)
                    log.info('old file deleted: ' + str(file))
            self._model = new_model(self.epc_file)

        return self._model

    def __exit__(self, exc_type, exc_value, exc_tb):
        """Exit the runtime context, close the model."""

        # Only write to disk if no exception has occured
        if self.mode in ["read/write", "create"] and exc_type is None:
            self._model.store_epc()

        # Release file handles
        self._model.h5_release()



"""Array writing functions."""

# Nexus is a registered trademark of the Halliburton Company

import logging

log = logging.getLogger(__name__)

# only nexus format currently fully supported

import math as maths
import numpy as np

import resqpy.olio.ab_toolbox as abt


def write_pure_binary_data(binary_file_name, numpy_array):
    """Writes a numpy array to a file in 'pure binary' format."""

    with open(binary_file_name, 'wb') as binary_file_out:
        numpy_array.tofile(binary_file_out)
    log.info('Binary data file %s created', binary_file_name)


def write_array_to_ascii_file(file_name,
                              extent_kji,
                              a,
                              headers = True,
                              keyword = None,
                              columns = 20,
                              data_type = 'real',
                              decimals = 3,
                              target_simulator = 'nexus',
                              blank_line_after_i_block = True,
                              blank_line_after_j_block = False,
                              space_separated = False,
                              append = False,
                              use_binary = False,
                              binary_only = False,
                              nan_substitute_value = None):
    """Writes a 3D array of data to an ascii file."""

    assert columns > 0 and decimals >= 0  # todo: test behaviour when decimals = 0
    assert target_simulator == 'nexus'  # no other simulator formats supported at the moment

    if not (use_binary and binary_only):

        format_str = ''
        if data_type == 'real' or data_type == 'float':
            format_str = '{0:.' + str(decimals) + 'f}'

        try:

            if append:
                file_mode = 'a'
            else:
                file_mode = 'w'

            with open(file_name, file_mode) as new_file:

                if headers:  # write 3 comment lines at start of file
                    new_file.write('! Data written by write_array_to_ascii_file() python function\n')
                    new_file.write('! Extent of array is: [' + str(extent_kji[2]) + ', ' + str(extent_kji[1]) + ', ' +
                                   str(extent_kji[0]) + ']\n')
                    new_file.write('! Maximum ' + str(columns) + ' data items per line\n')

                if keyword is not None:
                    new_file.write(keyword + '\n')

                for k in range(extent_kji[0]):
                    for j in range(extent_kji[1]):
                        for i in range(extent_kji[2]):
                            if (i % columns == 0):
                                new_file.write('\n')
                            elif space_separated:
                                new_file.write(' ')
                            else:
                                new_file.write('\t')
                            if data_type == 'real' or data_type == 'float':  # todo: speed up by using common write, different format
                                v = a[k, j, i]
                                if nan_substitute_value is not None and maths.isnan(v):
                                    v = nan_substitute_value
                                new_file.write(format_str.format(v))
                            elif data_type == 'bool' or data_type == 'boolean':
                                if a[k, j, i]:
                                    new_file.write('1')
                                else:
                                    new_file.write('0')
                            else:
                                new_file.write(str(a[k, j, i]))  # todo: other formatting
                        if blank_line_after_i_block:
                            new_file.write('\n')
                    if blank_line_after_j_block:
                        new_file.write('\n')

                if not (blank_line_after_i_block or blank_line_after_j_block):
                    new_file.write('\n')

            log.info('Ascii data file %s created', file_name)

        except Exception:
            log.error('Failed to write data to ascii file %s', file_name)
            # could abort at this point, or raise

    if use_binary:
        extension, _ = abt.binary_file_extension_and_np_type_for_data_type(data_type)
        binary_file_name = file_name + extension
        if append and (keyword is not None):
            key = keyword.split()[0].lower()
            if key != 'corp':
                binary_file_name = file_name + '_' + key + extension
        if nan_substitute_value is None:
            ap = a
        else:
            ap = np.where(np.isnan(a), nan_substitute_value, a)
        try:
            with open(binary_file_name, 'wb') as binary_file_out:
                ap.tofile(binary_file_out)
            log.info('Binary data file %s created', binary_file_name)
        except Exception:
            log.warning('Failed to write data to binary file %s', binary_file_name)
            # todo: could delete the binary file in case a corrupt file is left for use next time



"""_catalogue.py: RESQML parts (high level objects) catalogue functions."""

import logging

log = logging.getLogger(__name__)

import zipfile as zf

import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet


def _parts(model,
           parts_list = None,
           obj_type = None,
           uuid = None,
           title = None,
           title_mode = 'is',
           title_case_sensitive = False,
           metadata = {},
           extra = {},
           related_uuid = None,
           epc_subdir = None,
           sort_by = None):
    """Returns a list of parts matching all of the arguments passed."""

    if not parts_list:
        parts_list = _list_of_parts(model)
    if uuid is not None:
        part_name = model.uuid_part_dict.get(bu.uuid_as_int(uuid))
        if part_name is None or part_name not in parts_list:
            return []
        parts_list = [part_name]
    if epc_subdir:
        parts_list = _filtered_by_epc_subdir(model, parts_list, epc_subdir)
    if obj_type:
        if obj_type[0].isupper():
            obj_type = 'obj_' + obj_type
        filtered_list = []
        for part in parts_list:
            if model.parts_forest[part][0] == obj_type:
                filtered_list.append(part)
        if len(filtered_list) == 0:
            return []
        parts_list = filtered_list
    if title:
        parts_list = _filtered_by_title(model, parts_list, title, title_mode, title_case_sensitive)
    if metadata:
        parts_list = _filtered_by_metadata(model, parts_list, metadata)
    if extra:
        parts_list = _filtered_by_extra(model, parts_list, extra)
    if related_uuid is not None:
        parts_list = _parts_list_filtered_by_related_uuid(model, parts_list, related_uuid)
    if sort_by and len(parts_list):
        parts_list = _sorted_parts_list(model, parts_list, sort_by)
    return parts_list


def _part(model,
          parts_list = None,
          obj_type = None,
          uuid = None,
          title = None,
          title_mode = 'is',
          title_case_sensitive = False,
          metadata = {},
          extra = {},
          related_uuid = None,
          epc_subdir = None,
          multiple_handling = 'exception'):
    """Returns the name of a part matching all of the arguments passed."""

    pl = _parts(model,
                parts_list = parts_list,
                obj_type = obj_type,
                uuid = uuid,
                title = title,
                title_mode = title_mode,
                title_case_sensitive = title_case_sensitive,
                metadata = metadata,
                extra = extra,
                related_uuid = related_uuid,
                epc_subdir = epc_subdir)
    if len(pl) == 0:
        return None
    if len(pl) == 1 or multiple_handling == 'first':
        return pl[0]
    if multiple_handling == 'none':
        return None
    elif multiple_handling in ['newest', 'oldest']:
        sorted_list = _sort_parts_list_by_timestamp(model, pl)
        if multiple_handling == 'newest':
            return sorted_list[0]
        return sorted_list[-1]
    else:
        raise ValueError('more than one part matches criteria')


def _uuids(model,
           parts_list = None,
           obj_type = None,
           uuid = None,
           title = None,
           title_mode = 'is',
           title_case_sensitive = False,
           metadata = {},
           extra = {},
           related_uuid = None,
           epc_subdir = None,
           sort_by = None):
    """Returns a list of uuids of parts matching all of the arguments passed."""

    sort_by_uuid = (sort_by == 'uuid')
    if sort_by_uuid:
        sort_by = None
    pl = _parts(model,
                parts_list = parts_list,
                obj_type = obj_type,
                uuid = uuid,
                title = title,
                title_mode = title_mode,
                title_case_sensitive = title_case_sensitive,
                metadata = metadata,
                extra = extra,
                related_uuid = related_uuid,
                epc_subdir = epc_subdir,
                sort_by = sort_by)
    if len(pl) == 0:
        return []
    uuid_list = []
    for part in pl:
        uuid_list.append(_uuid_for_part(model, part))
    if sort_by_uuid:
        uuid_list.sort()
    return uuid_list


def _uuid(model,
          parts_list = None,
          obj_type = None,
          uuid = None,
          title = None,
          title_mode = 'is',
          title_case_sensitive = False,
          metadata = {},
          extra = {},
          related_uuid = None,
          epc_subdir = None,
          multiple_handling = 'exception'):
    """Returns the uuid of a part matching all of the arguments passed."""

    part = _part(model,
                 parts_list = parts_list,
                 obj_type = obj_type,
                 uuid = uuid,
                 title = title,
                 title_mode = title_mode,
                 title_case_sensitive = title_case_sensitive,
                 metadata = metadata,
                 extra = extra,
                 related_uuid = related_uuid,
                 epc_subdir = epc_subdir,
                 multiple_handling = multiple_handling)
    if part is None:
        return None
    return rqet.uuid_in_part_name(part)


def _roots(model,
           parts_list = None,
           obj_type = None,
           uuid = None,
           title = None,
           title_mode = 'is',
           title_case_sensitive = False,
           metadata = {},
           extra = {},
           related_uuid = None,
           epc_subdir = None,
           sort_by = None):
    """Returns a list of xml root nodes of parts matching all of the arguments passed."""

    pl = _parts(model,
                parts_list = parts_list,
                obj_type = obj_type,
                uuid = uuid,
                title = title,
                title_mode = title_mode,
                title_case_sensitive = title_case_sensitive,
                metadata = metadata,
                extra = extra,
                related_uuid = related_uuid,
                epc_subdir = epc_subdir,
                sort_by = sort_by)
    root_list = []
    for part in pl:
        root_list.append(_root_for_part(model, part))
    return root_list


def _root(model,
          parts_list = None,
          obj_type = None,
          uuid = None,
          title = None,
          title_mode = 'is',
          title_case_sensitive = False,
          metadata = {},
          extra = {},
          related_uuid = None,
          epc_subdir = None,
          multiple_handling = 'exception'):
    """Returns the xml root node of a part matching all of the arguments passed."""

    part = _part(model,
                 parts_list = parts_list,
                 obj_type = obj_type,
                 uuid = uuid,
                 title = title,
                 title_mode = title_mode,
                 title_case_sensitive = title_case_sensitive,
                 metadata = metadata,
                 extra = extra,
                 related_uuid = related_uuid,
                 epc_subdir = epc_subdir,
                 multiple_handling = multiple_handling)
    if part is None:
        return None
    return _root_for_part(model, part)


def _titles(model,
            parts_list = None,
            obj_type = None,
            uuid = None,
            title = None,
            title_mode = 'is',
            title_case_sensitive = False,
            metadata = {},
            extra = {},
            related_uuid = None,
            epc_subdir = None,
            sort_by = None):
    """Returns a list of citation titles of parts matching all of the arguments passed."""

    pl = _parts(model,
                parts_list = parts_list,
                obj_type = obj_type,
                uuid = uuid,
                title = title,
                title_mode = title_mode,
                title_case_sensitive = title_case_sensitive,
                metadata = metadata,
                extra = extra,
                related_uuid = related_uuid,
                epc_subdir = epc_subdir,
                sort_by = sort_by)
    title_list = []
    for part in pl:
        title_list.append(_citation_title_for_part(model, part))
    return title_list


def _title(model,
           parts_list = None,
           obj_type = None,
           uuid = None,
           title = None,
           title_mode = 'is',
           title_case_sensitive = False,
           metadata = {},
           extra = {},
           related_uuid = None,
           epc_subdir = None,
           multiple_handling = 'exception'):
    """Returns the citation title of a part matching all of the arguments passed."""

    part = _part(model,
                 parts_list = parts_list,
                 obj_type = obj_type,
                 uuid = uuid,
                 title = title,
                 title_mode = title_mode,
                 title_case_sensitive = title_case_sensitive,
                 metadata = metadata,
                 extra = extra,
                 related_uuid = related_uuid,
                 epc_subdir = epc_subdir,
                 multiple_handling = multiple_handling)
    if part is None:
        return None
    return _citation_title_for_part(model, part)


def _parts_list_of_type(model, type_of_interest = None, uuid = None):
    """Returns a list of part names for parts of type of interest, optionally matching a uuid."""

    if type_of_interest and type_of_interest[0].isupper():
        type_of_interest = 'obj_' + type_of_interest

    if uuid is not None:
        part_name = model.uuid_part_dict.get(bu.uuid_as_int(uuid))
        if part_name is None or (type_of_interest is not None and
                                 (model.parts_forest[part_name][0] != type_of_interest)):
            return []
        return [part_name]

    parts_list = []
    for part_name in model.parts_forest:
        if type_of_interest is None or model.parts_forest[part_name][0] == type_of_interest:
            parts_list.append(part_name)
    return parts_list


def _list_of_parts(model, only_objects = True):
    """Return a complete list of parts."""

    pl = list(model.parts_forest.keys())
    if not only_objects:
        return pl
    obj_list = []
    for part in pl:
        dir_place = part.rfind('/')
        dir_free_part = part[dir_place + 1:]
        if dir_free_part.startswith('obj_') and not dir_free_part.startswith('obj_Epc'):
            obj_list.append(part)
    return obj_list


def _number_of_parts(model):
    """Retuns the number of parts in the model, including external parts such as the link to an hdf5 file."""

    return len(model.parts_forest)


def _part_for_uuid(model, uuid):
    """Returns the part name which has the given uuid."""

    return model.uuid_part_dict.get(bu.uuid_as_int(uuid))


def _root_for_uuid(model, uuid):
    """Returns the xml root for the part which has the given uuid."""

    return _root_for_part(model, _part_for_uuid(model, uuid))


def _parts_count_by_type(model, type_of_interest = None):
    """Returns a sorted list of (type, count) for parts."""

    # note: resqml classes start with 'obj_' whilst witsml classes don't!
    if type_of_interest and type_of_interest.startswith('obj_'):
        type_of_interest = type_of_interest[4:]

    type_list = []
    for part_name in model.parts_forest:
        part_type = model.parts_forest[part_name][0]
        if part_type is None:
            continue
        if part_type.startswith('obj_'):
            part_type = part_type[4:]
        if type_of_interest is None or part_type == type_of_interest:
            type_list.append(part_type)
    type_list.sort()
    type_list.append('END')  # simplifies termination of scan below
    result_list = []
    count = 0
    current_type = ''
    for index in range(len(type_list)):
        if type_list[index] != current_type:
            if count:
                result_list.append((current_type, count))
            current_type = type_list[index]
            count = 0
        count += 1
    return result_list


def _parts_list_filtered_by_related_uuid(model, parts_list, uuid, uuid_is_source = None):
    """From a list of parts, returns a list of those parts which have a relationship with the given uuid."""

    if not model.rels_present or parts_list is None or uuid is None:
        return None
    filtered_list = []
    this_part = _part_for_uuid(model, uuid)

    if this_part is not None:
        rels_part_root = _root_for_part(model, rqet.rels_part_name_for_part(this_part), is_rels = True)
        if rels_part_root is not None:
            for relation_node in rels_part_root:
                if rqet.stripped_of_prefix(relation_node.tag) != 'Relationship':
                    continue
                target_part = relation_node.attrib['Target']
                if target_part not in parts_list:
                    continue
                if uuid_is_source is not None:
                    source_dest = relation_node.attrib['Type']
                    if uuid_is_source:
                        if 'source' not in source_dest:
                            continue
                    else:
                        if 'source' in source_dest:
                            continue
                filtered_list.append(target_part)

    for part in parts_list:
        if part in filtered_list:
            continue
        rels_part_root = _root_for_part(model, rqet.rels_part_name_for_part(part), is_rels = True)
        if rels_part_root is None:
            continue
        for relation_node in rels_part_root:
            if rqet.stripped_of_prefix(relation_node.tag) != 'Relationship':
                continue
            target_part = relation_node.attrib['Target']
            relation_uuid = rqet.uuid_in_part_name(target_part)
            if bu.matching_uuids(uuid, relation_uuid):
                if uuid_is_source is not None:
                    source_dest = relation_node.attrib['Type']
                    if uuid_is_source:
                        if 'source' in source_dest:
                            continue  # relation is source, so uuid is not
                    else:
                        if 'source' not in source_dest:
                            continue  # relation is not source, so uuid is
                filtered_list.append(part)
                break

    return filtered_list


def _supporting_representation_for_part(model, part):
    """Returns the uuid of the supporting representation for the part, if found, otherwise None."""

    return bu.uuid_from_string(
        rqet.find_nested_tags_text(_root_for_part(model, part), ['SupportingRepresentation', 'UUID']))


def _parts_list_filtered_by_supporting_uuid(model, parts_list, uuid):
    """From a list of parts, returns a list of those parts which have the given uuid as supporting representation."""

    if parts_list is None or uuid is None:
        return None
    filtered_list = []
    for part in parts_list:
        support_ref_uuid = _supporting_representation_for_part(model, part)
        if support_ref_uuid is None:
            continue
        if bu.matching_uuids(support_ref_uuid, uuid):
            filtered_list.append(part)
    return filtered_list


def _parts_list_related_to_uuid_of_type(model, uuid, type_of_interest = None):
    """Returns a list of parts of type of interest that relate to part with given uuid."""

    parts_list = _parts_list_of_type(model, type_of_interest = type_of_interest)
    return _parts_list_filtered_by_related_uuid(model, parts_list, uuid)


def _external_parts_list(model):
    """Returns a list of part names for external part references."""

    return _parts_list_of_type(model, 'obj_EpcExternalPartReference')


def _uuid_for_part(model, part_name, is_rels = None):
    """Returns the uuid for the named part."""

    if part_name is None:
        return None
    if is_rels is None:
        is_rels = part_name.endswith('.rels')
    if is_rels:
        return model.rels_forest[part_name][0]
    return model.parts_forest[part_name][1]


def _type_of_part(model, part_name, strip_obj = False):
    """Returns content type for the named part (does not apply to rels parts)."""

    part_info = model.parts_forest.get(part_name)
    if part_info is None:
        return None
    obj_type = part_info[0]
    if obj_type is None or not strip_obj or not obj_type.startswith('obj_'):
        return obj_type
    return obj_type[4:]


def _type_of_uuid(model, uuid, strip_obj = False):
    """Returns content type for the uuid."""

    part_name = model.uuid_part_dict.get(bu.uuid_as_int(uuid))
    return _type_of_part(model, part_name, strip_obj = strip_obj)


def _tree_for_part(model, part_name, is_rels = None):
    """Returns parsed xml tree for the named part."""

    if not part_name:
        return None
    if is_rels is None:
        is_rels = part_name.endswith('.rels')
    is_other = not is_rels and part_name.startswith('docProps')
    if is_rels:
        if part_name not in model.rels_forest:
            return None
        (_, tree) = model.rels_forest[part_name]
        if tree is None:
            if not model.epc_file:
                return None
            with zf.ZipFile(model.epc_file) as epc:
                load_success = model.load_part(epc, part_name, is_rels = True)
                if not load_success:
                    return None
        return model.rels_forest[part_name][1]
    elif is_other:
        if part_name not in model.other_forest:
            return None
        (_, tree) = model.other_forest[part_name]
        if tree is None:
            if not model.epc_file:
                return None
            with zf.ZipFile(model.epc_file) as epc:
                load_success = model.load_part(epc, part_name, is_rels = False)
                if not load_success:
                    return None
        return model.other_forest[part_name][1]
    else:
        if part_name not in model.parts_forest:
            return None
        (_, _, tree) = model.parts_forest[part_name]
        if tree is None:
            if not model.epc_file:
                return None
            with zf.ZipFile(model.epc_file) as epc:
                load_success = model.load_part(epc, part_name, is_rels = False)
                if not load_success:
                    return None
        return model.parts_forest[part_name][2]


def _root_for_part(model, part_name, is_rels = None):
    """Returns root of parsed xml tree for the named part."""

    if not part_name:
        return None
    tree = _tree_for_part(model, part_name, is_rels = is_rels)
    if tree is None:
        return None
    return tree.getroot()


def _citation_title_for_part(model, part):  # duplicate functionality to title_for_part()
    """Returns the citation title for the specified part."""

    return rqet.citation_title_for_node(_root_for_part(model, part))


def _root_for_time_series(model, uuid = None):
    """Return root for time series part."""

    time_series_list = _parts_list_of_type(model, 'obj_TimeSeries', uuid = uuid)
    if len(time_series_list) == 0:
        return None
    if len(time_series_list) == 1:
        return _root_for_part(model, time_series_list[0])
    log.warning('selecting time series with earliest creation date')
    oldest_root = oldest_creation = None
    for ts in time_series_list:
        node = _root_for_part(model, ts)
        created = rqet.creation_date_for_node(node)
        if oldest_creation is None or created < oldest_creation:
            oldest_creation = created
            oldest_root = node
    return oldest_root


def _resolve_time_series_root(model, time_series_root = None):
    """If time_series_root is None, finds the root for a time series in the model."""

    if time_series_root is not None:
        return time_series_root
    if model.time_series is None:
        model.time_series = _root_for_time_series(model)
    return model.time_series


def _title_for_root(model, root = None):
    """Returns the Title text from the Citation within the given root node."""

    title = rqet.find_tag(rqet.find_tag(root, 'Citation'), 'Title')
    if title is None:
        return None

    return title.text


def _title_for_part(model, part_name):  # duplicate functionality to citation_title_for_part()
    """Returns the Title text from the Citation for the given main part name (not for rels)."""

    return _title_for_root(model, _root_for_part(model, part_name))


def _iter_objs(model, cls):
    """Iterate over all available objects of given resqpy class within the model."""

    uuids = _uuids(model, obj_type = cls.resqml_type)
    for uuid in uuids:
        yield cls(model, uuid = uuid)


def _iter_grid_connection_sets(model):
    """Yields grid connection set objects, one for each gcs in this model."""

    import resqpy.fault as rqf  # imported here for speed, module is not always needed

    gcs_uuids = _uuids(model, obj_type = 'GridConnectionSetRepresentation')
    for gcs_uuid in gcs_uuids:
        yield rqf.GridConnectionSet(model, uuid = gcs_uuid)


def _iter_wellbore_interpretations(model):
    """Iterable of all WellboreInterpretations associated with the model."""

    import resqpy.organize as rqo  # imported here for speed, module is not always needed

    uuids = _uuids(model, obj_type = 'WellboreInterpretation')
    if uuids:
        for uuid in uuids:
            yield rqo.WellboreInterpretation(model, uuid = uuid)


def _iter_trajectories(model):
    """Iterable of all trajectories associated with the model."""

    import resqpy.well as rqw  # imported here for speed, module is not always needed

    uuids = _uuids(model, obj_type = "WellboreTrajectoryRepresentation")
    for uuid in uuids:
        yield rqw.Trajectory(model, uuid = uuid)


def _iter_md_datums(model):
    """Iterable of all MdDatum objects associated with the model."""

    import resqpy.well as rqw  # imported here for speed, module is not always needed

    uuids = _uuids(model, obj_type = 'MdDatum')
    if uuids:
        for uuid in uuids:
            datum = rqw.MdDatum(model, uuid = uuid)
            yield datum


def _iter_crs(model):
    """Iterable of all CRS objects associated with the model."""

    import resqpy.crs as rqc  # imported here for speed, module is not always needed

    uuids = _uuids(model, obj_type = 'LocalDepth3dCrs') + _uuids(model, obj_type = 'LocalTime3dCrs')
    if uuids:
        for uuid in uuids:
            yield rqc.Crs(model, uuid = uuid)


def _sort_parts_list_by_timestamp(model, parts_list):
    """Returns a copy of the parts list sorted by citation block creation date, with the newest first."""

    if parts_list is None:
        return None
    if len(parts_list) == 0:
        return []
    sort_list = []
    for index, part in enumerate(parts_list):
        timestamp = rqet.find_nested_tags_text(_root_for_part(model, part), ['Citation', 'Creation'])
        sort_list.append((timestamp, index))
    sort_list.sort()
    results = []
    for timestamp, index in reversed(sort_list):
        results.append(parts_list[index])
    return results


def _as_graph(model, uuids_subset = None):
    """Return representation of model as nodes and edges, suitable for plotting in a graph."""

    nodes = {}
    edges = set()

    if uuids_subset is None:
        uuids_subset = _uuids(model)

    uuids_subset = set(map(str, uuids_subset))

    for uuid in uuids_subset:
        part = _part_for_uuid(model, uuid)
        nodes[uuid] = dict(
            resqml_type = _type_of_part(model, part, strip_obj = True),
            title = _citation_title_for_part(model, part),
        )
        for rel in map(str, _uuids(model, related_uuid = uuid)):
            if rel in uuids_subset:
                edges.add(frozenset([uuid, rel]))

    return nodes, edges


def _filtered_by_epc_subdir(model, parts_list, epc_subdir):
    if epc_subdir.startswith('/'):
        epc_subdir = epc_subdir[1:]
    if epc_subdir:
        if not epc_subdir.endswith('/'):
            epc_subdir += '/'
        filtered_list = []
        for part in parts_list:
            if part.startswith[epc_subdir]:
                filtered_list.append(part)
        return filtered_list
    else:
        return parts_list


def _filtered_by_title(model, parts_list, title, title_mode, title_case_sensitive):
    assert title_mode in [
        'is', 'starts', 'ends', 'contains', 'is not', 'does not start', 'does not end', 'does not contain'
    ]
    if not title_case_sensitive:
        title = title.upper()
    filtered_list = []
    for part in parts_list:
        part_title = _citation_title_for_part(model, part)
        if not title_case_sensitive:
            part_title = part_title.upper()
        if title_mode == 'is':
            if part_title == title:
                filtered_list.append(part)
        elif title_mode == 'starts':
            if part_title.startswith(title):
                filtered_list.append(part)
        elif title_mode == 'ends':
            if part_title.endswith(title):
                filtered_list.append(part)
        elif title_mode == 'contains':
            if title in part_title:
                filtered_list.append(part)
        if title_mode == 'is not':
            if part_title != title:
                filtered_list.append(part)
        elif title_mode == 'does not start':
            if not part_title.startswith(title):
                filtered_list.append(part)
        elif title_mode == 'does not end':
            if not part_title.endswith(title):
                filtered_list.append(part)
        elif title_mode == 'does not contain':
            if title not in part_title:
                filtered_list.append(part)
    return filtered_list


def _filtered_by_metadata(model, parts_list, metadata):
    filtered_list = []
    for part in parts_list:
        root = _root_for_part(model, part)
        match = True
        for key, value in metadata.items():
            node = rqet.find_tag(root, str(key))
            if node is None or node.text != str(value):
                match = False
                break
        if match:
            filtered_list.append(part)
    return filtered_list


def _filtered_by_extra(model, parts_list, extra):
    filtered_list = []
    for part in parts_list:
        part_extra = rqet.load_metadata_from_xml(_root_for_part(model, part))
        if not part_extra:
            continue
        match = True
        for key, value in extra.items():
            if key not in part_extra or part_extra[key] != value:
                match = False
                break
        if match:
            filtered_list.append(part)
    return filtered_list


def _sorted_parts_list(model, parts_list, sort_by):
    if sort_by == 'type':
        sorted_list = sorted(parts_list)
    elif sort_by in ['newest', 'oldest']:
        sorted_list = _sort_parts_list_by_timestamp(model, parts_list)
        if sort_by == 'oldest':
            sorted_list.reverse()
    elif sort_by in ['uuid', 'title']:
        sort_list = []
        for index, part in enumerate(parts_list):
            if sort_by == 'uuid':
                key = str(_uuid_for_part(model, part))
            else:
                key = _citation_title_for_part(model, part)
            sort_list.append((key, index))
        sort_list.sort()
        sorted_list = []
        for _, index in sort_list:
            sorted_list.append(parts_list[index])
    return sorted_list



"""Class for RESQML Geobody Interpretation objects."""

import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
import resqpy.organize
import resqpy.organize.geobody_feature as ogf
import resqpy.organize._utils as ou
from resqpy.olio.base import BaseResqpy
from resqpy.olio.xml_namespaces import curly_namespace as ns


class GeobodyInterpretation(BaseResqpy):
    """Class for RESQML Geobody Interpretation objects."""

    resqml_type = 'GeobodyInterpretation'
    valid_domains = ('depth', 'time', 'mixed')
    valid_compositions = (
        'intrusive clay',
        'organic',
        'intrusive mud',
        'evaporite salt',
        'evaporite non salt',
        'sedimentary siliclastic',
        'carbonate',
        'magmatic intrusive granitoid',
        'magmatic intrusive pyroclastic',
        'magmatic extrusive lava flow',
        'other chemichal rock',  # chemichal (stet: from xsd)
        'other chemical rock',
        'sedimentary turbidite')
    valid_implacements = ('autochtonous', 'allochtonous')
    valid_geobody_shapes = ('dyke', 'silt', 'sill', 'dome', 'sheeth', 'sheet', 'diapir', 'batholith', 'channel',
                            'delta', 'dune', 'fan', 'reef', 'wedge')

    def __init__(self,
                 parent_model,
                 uuid = None,
                 title = None,
                 geobody_feature = None,
                 domain = 'depth',
                 composition = None,
                 material_implacement = None,
                 geobody_shape = None,
                 extra_metadata = None):
        """Initialise a new geobody interpretation object, either from xml or explicitly."""

        self.domain = domain
        self.geobody_feature = geobody_feature  # InterpretedFeature RESQML field, when not loading from xml
        self.feature_root = None if self.geobody_feature is None else self.geobody_feature.root
        self.has_occurred_during = (None, None)
        self.composition = composition
        self.implacement = material_implacement
        self.geobody_shape = geobody_shape
        super().__init__(model = parent_model, uuid = uuid, title = title, extra_metadata = extra_metadata)

    def _load_from_xml(self):
        root_node = self.root
        interp_feature_ref_node = rqet.find_tag(root_node, 'InterpretedFeature')
        assert interp_feature_ref_node is not None
        self.feature_root = self.model.referenced_node(interp_feature_ref_node)
        if self.feature_root is not None:
            self.geobody_feature = ogf.GeobodyFeature(self.model,
                                                      uuid = self.feature_root.attrib['uuid'],
                                                      feature_name = self.model.title_for_root(self.feature_root))
        self.has_occurred_during = ou.extract_has_occurred_during(root_node)
        self.composition = rqet.find_tag_text(root_node, 'GeologicUnitComposition')
        self.implacement = rqet.find_tag_text(root_node, 'GeologicUnitMaterialImplacement')
        self.geobody_shape = rqet.find_tag_text(root_node, 'Geobody3dShape')

    def is_equivalent(self, other, check_extra_metadata = True):
        """Returns True if this interpretation is essentially the same as the other; otherwise False."""

        if other is None or not isinstance(other, GeobodyInterpretation):
            return False
        if self is other or bu.matching_uuids(self.uuid, other.uuid):
            return True
        if self.geobody_feature is not None:
            if not self.geobody_feature.is_equivalent(other.geobody_feature,
                                                      check_extra_metadata = check_extra_metadata):
                return False
        elif other.geobody_feature is not None:
            return False
        if self.root is not None and other.root is not None:
            if rqet.citation_title_for_node(self.root) != rqet.citation_title_for_node(other.root):
                return False
        elif self.root is not None or other.root is not None:
            return False
        if check_extra_metadata and not ou.equivalent_extra_metadata(self, other):
            return False
        return (self.domain == other.domain and
                ou.equivalent_chrono_pairs(self.has_occurred_during, other.has_occurred_during) and
                self.composition == other.composition and self.implacement == other.implacement and
                self.geobody_shape == other.geobody_shape)

    def create_xml(self,
                   geobody_feature_root = None,
                   add_as_part = True,
                   add_relationships = True,
                   originator = None,
                   title_suffix = None,
                   reuse = True):
        """Creates an XML tree in memory and optionally adds it as a 'part' in the Model"""

        if not self.title:
            self.title = self.geobody_feature.feature_name
        if title_suffix:
            self.title += ' ' + title_suffix

        if reuse and self.try_reuse():
            return self.root
        gi = super().create_xml(add_as_part = False, originator = originator)

        if self.geobody_feature is not None:
            gbf_root = self.geobody_feature.root
            if gbf_root is not None:
                if geobody_feature_root is None:
                    geobody_feature_root = gbf_root
                else:
                    assert gbf_root is geobody_feature_root, 'geobody feature mismatch'
        else:
            if geobody_feature_root is None:
                geobody_feature_root = self.feature_root
            assert geobody_feature_root is not None
            self.geobody_feature = ogf.GeobodyFeature(self.model, uuid = geobody_feature_root.attrib['uuid'])
        self.feature_root = geobody_feature_root

        assert self.domain in self.valid_domains, 'illegal domain value for geobody interpretation'
        dom_node = rqet.SubElement(gi, ns['resqml2'] + 'Domain')
        dom_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Domain')
        dom_node.text = self.domain

        ou.create_xml_has_occurred_during(self.model, gi, self.has_occurred_during)

        if self.composition:
            assert self.composition in self.valid_compositions
            guc_node = rqet.SubElement(gi, ns['resqml2'] + 'GeologicUnitComposition')
            guc_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'GeologicUnitComposition')
            guc_node.text = self.composition
            # if self.composition.startswith('intrusive'): guc_node.text += ' '

        if self.implacement:
            assert self.implacement in self.valid_implacements
            gumi_node = rqet.SubElement(gi, ns['resqml2'] + 'GeologicUnitMaterialImplacement')
            gumi_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'GeologicUnitMaterialImplacement')
            gumi_node.text = self.implacement

        if self.geobody_shape:
            # note: 'silt' & 'sheeth' believed erroneous, so 'sill' and 'sheet' added
            assert self.geobody_shape in self.valid_geobody_shapes
            gs_node = rqet.SubElement(gi, ns['resqml2'] + 'Geobody3dShape')
            gs_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Geobody3dShape')
            gs_node.text = self.geobody_shape

        self.model.create_ref_node('InterpretedFeature',
                                   self.model.title_for_root(geobody_feature_root),
                                   geobody_feature_root.attrib['uuid'],
                                   content_type = 'obj_GeobodyFeature',
                                   root = gi)

        if add_as_part:
            self.model.add_part('obj_GeobodyInterpretation', self.uuid, gi)
            if add_relationships:
                self.model.create_reciprocal_relationship(gi, 'destinationObject', geobody_feature_root, 'sourceObject')

        return gi



"""Submodule containing some general grid functions that have now been moved. This file will be deleted in a later release."""

import warnings

import resqpy.grid._points_functions as pf
import resqpy.property.property_kind as pk


def establish_zone_property_kind(model):
    """MOVED: Returns zone local property kind object, creating the xml and adding as part if not found in model."""
    warnings.warn(
        'This function has been moved to property/property_kind. Please update your code to the new '
        'location.', DeprecationWarning)

    return pk.establish_zone_property_kind(model)



"""High level interpolated grid function."""

import logging

log = logging.getLogger(__name__)

import os
import numpy as np

import resqpy.crs as rqc
import resqpy.derived_model
import resqpy.grid as grr
import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet

import resqpy.derived_model._common as rqdm_c


def interpolated_grid(epc_file,
                      grid_a,
                      grid_b,
                      a_to_b_0_to_1 = 0.5,
                      split_tolerance = 0.01,
                      inherit_properties = False,
                      inherit_realization = None,
                      inherit_all_realizations = False,
                      new_grid_title = None,
                      new_epc_file = None):
    """Extends an existing model with a new grid geometry linearly interpolated between the two source_grids.

    arguments:
       epc_file (string): file name to rewrite the model's xml to
       grid_a, grid_b (grid.Grid objects): a pair of RESQML grid objects representing the end cases, between
          which the new grid will be interpolated
       a_to_b_0_to_1 (float, default 0.5): the interpolation factor in the range zero to one; a value of 0.0 will yield
          a copy of grid a, a value of 1.0 will yield a copy of grid b, intermediate values will yield a grid with all
          points interpolated
       split_tolerance (float, default 0.01): maximum offset of corner points for shared point to be generated; units
          are same as those in grid crs; only relevant if working from corner points, ignored otherwise
       inherit_properties (boolean, default False): if True, the new grid will have a copy of any properties associated
          with grid_a
       inherit_realization (int, optional): realization number for which properties will be inherited; ignored if
          inherit_properties is False
       inherit_all_realizations (boolean, default False): if True (and inherit_realization is None), properties for all
          realizations will be inherited; if False, only properties with a realization of None are inherited; ignored if
          inherit_properties is False or inherit_realization is not None
       new_grid_title (string): used as the citation title text for the new grid object
       new_epc_file (string, optional): if None, the source epc_file is extended with the new grid object; if present,
          a new epc file (& associated h5 file) is created to contain the interpolated grid (& crs)

    returns:
       new grid object (grid.Grid) with geometry interpolated between grid a and grid b

    notes:
       the hdf5 file used by the grid_a model is appended to, so it is recommended that the grid_a model's epc is specified
       as the first argument (unless a new epc file is required, sharing the hdf5 file)
    """

    assert epc_file or new_epc_file, 'epc file name not specified'
    if new_epc_file and epc_file and (
        (new_epc_file == epc_file) or
        (os.path.exists(new_epc_file) and os.path.exists(epc_file) and os.path.samefile(new_epc_file, epc_file))):
        new_epc_file = None
    assert grid_a is not None and grid_b is not None, 'at least one source grid is missing'
    assert grid_a.grid_representation == 'IjkGrid' and grid_b.grid_representation == 'IjkGrid'
    assert 0.0 <= a_to_b_0_to_1 <= 1.0, 'interpolation factor outside range 0.0 to 1.0'
    assert tuple(grid_a.extent_kji) == tuple(grid_b.extent_kji), 'source grids have different extents'
    assert grid_a.k_direction_is_down == grid_b.k_direction_is_down, 'source grids have different k directions'
    assert grid_a.grid_is_right_handed == grid_b.grid_is_right_handed, 'source grids have different ijk handedness'
    assert grid_a.pillar_shape == grid_b.pillar_shape, 'source grids have different resqml pillar shapes'

    b_weight = a_to_b_0_to_1
    a_weight = 1.0 - b_weight

    model = grid_a.model

    if not bu.matching_uuids(grid_a.crs_uuid, grid_b.crs_uuid):
        crs_a = rqc.Crs(grid_a.model, uuid = grid_a.crs_uuid)
        crs_b = rqc.Crs(grid_b.model, uuid = grid_b.crs_uuid)
        assert crs_a.is_equivalent(crs_b),  \
            'end point grids for interpolation have different coordinate reference systems'

    log.info('loading geometry for two source grids')
    grid_a.cache_all_geometry_arrays()
    grid_b.cache_all_geometry_arrays()

    assert (grid_a.geometry_defined_for_all_cells() and
            grid_b.geometry_defined_for_all_cells()), 'geometry not defined for all cells'
    # assert grid_a.geometry_defined_for_all_pillars() and grid_b.geometry_defined_for_all_pillars(),  \
    #     'geometry not defined for all pillars'

    work_from_pillars = _determine_work_from_pillars(grid_a, grid_b)

    # create a new, empty grid object
    grid = grr.Grid(model)

    # inherit attributes from source grid
    _inherit_basics(grid, grid_a, grid_b)

    if work_from_pillars:
        _interpolate_points_cached_from_pillars(grid, grid_a, grid_b, a_weight, b_weight)
    else:
        _interpolate_points_cached_from_cp(grid, grid_a, grid_b, a_weight, b_weight, split_tolerance)

    collection = rqdm_c._prepare_simple_inheritance(grid, grid_a, inherit_properties, inherit_realization,
                                                    inherit_all_realizations)

    if new_grid_title is None or len(new_grid_title) == 0:
        new_grid_title = 'interpolated between two grids with factor: ' + str(a_to_b_0_to_1)

    model.h5_release()
    if new_epc_file:
        rqdm_c._write_grid(new_epc_file,
                           grid,
                           property_collection = collection,
                           grid_title = new_grid_title,
                           mode = 'w')
    else:
        ext_uuid, _ = model.h5_uuid_and_path_for_node(rqet.find_nested_tags(grid_a.root, ['Geometry', 'Points']),
                                                      'Coordinates')
        rqdm_c._write_grid(epc_file,
                           grid,
                           ext_uuid = ext_uuid,
                           property_collection = collection,
                           grid_title = new_grid_title,
                           mode = 'a')

    return grid


def _determine_work_from_pillars(grid_a, grid_b):
    if not grid_a.has_split_coordinate_lines and not grid_b.has_split_coordinate_lines:
        work_from_pillars = True
    elif (grid_a.has_split_coordinate_lines and grid_b.has_split_coordinate_lines and
          grid_a.points_cached.shape == grid_b.points_cached.shape and
          grid_a.split_pillar_indices_cached.shape == grid_b.split_pillar_indices_cached.shape and
          grid_a.cols_for_split_pillars.shape == grid_b.cols_for_split_pillars.shape and
          grid_a.cols_for_split_pillars_cl.shape == grid_b.cols_for_split_pillars_cl.shape and
          np.all(grid_a.split_pillar_indices_cached == grid_b.split_pillar_indices_cached) and
          np.all(grid_a.cols_for_split_pillars == grid_b.cols_for_split_pillars) and
          np.all(grid_a.cols_for_split_pillars_cl == grid_b.cols_for_split_pillars_cl)):
        work_from_pillars = True
    else:
        work_from_pillars = False
    if work_from_pillars:
        log.info('interpolating between compatible pillar grids')
    else:
        log.warning('interpolating between corner points due to pillar incompatibilities')
    return work_from_pillars


def _inherit_basics(grid, grid_a, grid_b):
    grid.grid_representation = 'IjkGrid'
    grid.extent_kji = grid_a.extent_kji.copy()
    grid.nk, grid.nj, grid.ni = grid.extent_kji
    grid.k_direction_is_down = grid_a.k_direction_is_down
    grid.grid_is_right_handed = grid_a.grid_is_right_handed
    grid.pillar_shape = grid_a.pillar_shape
    grid.has_split_coordinate_lines = (grid_a.has_split_coordinate_lines or grid_b.has_split_coordinate_lines)
    # inherit the coordinate reference system used by the grid geometry
    grid.crs_uuid = grid_a.crs_uuid
    if grid_a.model is not grid.model:
        grid.model.duplicate_node(grid_a.model.root_for_uuid(grid_a, grid.crs_uuid), add_as_part = True)
    grid.crs = rqc.Crs(grid.model, uuid = grid.crs_uuid)

    if grid_a.inactive is None or grid_b.inactive is None:
        grid.inactive = None
    else:
        grid.inactive = np.logical_and(grid_a.inactive, grid_b.inactive)
    grid.geometry_defined_for_all_cells_cached = True
    grid.array_cell_geometry_is_defined = np.ones(tuple(grid.extent_kji), dtype = bool)
    grid.geometry_defined_for_all_pillars_cached = True


def _interpolate_points_cached_from_pillars(grid, grid_a, grid_b, a_weight, b_weight):
    grid.points_cached = grid_a.points_cached * a_weight + grid_b.points_cached * b_weight
    grid.has_split_coordinate_lines = grid_a.has_split_coordinate_lines
    if grid.has_split_coordinate_lines:
        grid.split_pillar_indices_cached = grid_a.split_pillar_indices_cached.copy()
        grid.cols_for_split_pillars = grid_a.cols_for_split_pillars.copy()
        grid.cols_for_split_pillars_cl = grid_a.cols_for_split_pillars_cl.copy()
        grid.split_pillars_count = grid_a.split_pillars_count


def _interpolate_points_cached_from_cp(grid, grid_a, grid_b, a_weight, b_weight, split_tolerance):
    grid.pillar_shape = 'curved'  # following fesapi approach of non-parametric pillars even if they are in fact straight
    cp_a = grid_a.corner_points(cache_cp_array = True)
    cp_b = grid_b.corner_points(cache_cp_array = True)
    assert cp_a.shape == cp_b.shape
    grid_cp = cp_a * a_weight + cp_b * b_weight
    _close_vertical_voids(grid, grid_cp, grid_a.z_units())
    # reduce cp array extent in k
    log.debug('reducing k extent of interpolated corner point array (sharing points vertically)')
    k_reduced_cp_array = np.zeros((grid.nk + 1, grid.nj, grid.ni, 2, 2, 3))  # (nk+1, nj, ni, jp, ip, xyz)
    k_reduced_cp_array[0, :, :, :, :, :] = grid_cp[0, :, :, 0, :, :, :]
    k_reduced_cp_array[-1, :, :, :, :, :] = grid_cp[-1, :, :, 1, :, :, :]
    if grid.nk > 1:
        k_reduced_cp_array[1:-1, :, :, :, :, :] = grid_cp[:-1, :, :, 1, :, :, :]
    # create primary pillar reference indices as one of four column corners around pillar, active column preferred
    log.debug('creating primary pillar reference neighbourly indices')
    primary_pillar_jip = np.zeros((grid.nj + 1, grid.ni + 1, 2), dtype = 'int')  # (nj + 1, ni + 1, jp:ip)
    primary_pillar_jip[-1, :, 0] = 1
    primary_pillar_jip[:, -1, 1] = 1
    # build extra pillar references for split pillars
    extras_count = np.zeros((grid.nj + 1, grid.ni + 1), dtype = 'int')  # count (0 to 3) of extras for pillar
    extras_list_index = np.zeros((grid.nj + 1, grid.ni + 1), dtype = 'int')  # index in list of 1st extra for pillar
    extras_list = []  # list of (jp, ip)
    # extras_use: (j, i, jp, ip); -1 means use primary
    extras_use = np.negative(np.ones((grid.nj, grid.ni, 2, 2), dtype = 'int'))
    log.debug('building extra pillar references for split pillars')
    # loop over pillars
    for j in range(grid.nj + 1):
        for i in range(grid.ni + 1):
            primary_jp = primary_pillar_jip[j, i, 0]
            primary_ip = primary_pillar_jip[j, i, 1]
            p_col_j = j - primary_jp
            p_col_i = i - primary_ip
            # loop over 4 columns surrounding this pillar
            for jp in range(2):
                col_j = j - jp
                if col_j < 0 or col_j >= grid.nj:
                    continue  # no column this side of pillar in j
                for ip in range(2):
                    col_i = i - ip
                    if col_i < 0 or col_i >= grid.ni:
                        continue  # no column this side of pillar in i
                    if jp == primary_jp and ip == primary_ip:
                        continue  # this column is the primary for this pillar
                    discrepancy = np.max(
                        np.abs(k_reduced_cp_array[:, col_j, col_i, jp, ip, :] -
                               k_reduced_cp_array[:, p_col_j, p_col_i, primary_jp, primary_ip, :]))
                    if discrepancy <= split_tolerance:
                        continue  # data for this column's corner aligns with primary
                    for e in range(extras_count[j, i]):
                        eli = extras_list_index[j, i] + e
                        pillar_j_extra = j - extras_list[eli][0]
                        pillar_i_extra = i - extras_list[eli][1]
                        discrepancy = np.max(
                            np.abs(k_reduced_cp_array[:, col_j, col_i, jp, ip, :] -
                                   k_reduced_cp_array[:, pillar_j_extra, pillar_i_extra, extras_list[eli][0],
                                                      extras_list[eli][1], :]))
                        if discrepancy <= split_tolerance:  # data for this corner aligns with existing extra
                            extras_use[col_j, col_i, jp, ip] = e
                            break
                    if extras_use[col_j, col_i, jp, ip] >= 0:  # reusing an existing extra for this pillar
                        continue
                    # add this corner as an extra
                    if extras_count[j, i] == 0:  # create entry point for this pillar in extras
                        extras_list_index[j, i] = len(extras_list)
                    extras_list.append((jp, ip))
                    extras_use[col_j, col_i, jp, ip] = extras_count[j, i]
                    extras_count[j, i] += 1
    if len(extras_list) == 0:
        grid.has_split_coordinate_lines = False
    log.debug('number of extra pillars: ' + str(len(extras_list)))
    # create points array as used in resqml
    log.debug('creating points array as used in resqml format')
    if grid.has_split_coordinate_lines:
        points_array = np.zeros((grid.nk + 1, (grid.nj + 1) * (grid.ni + 1) + len(extras_list), 3))
        index = 0
        # primary pillars
        for pillar_j in range(grid.nj + 1):
            for pillar_i in range(grid.ni + 1):
                (jp, ip) = primary_pillar_jip[pillar_j, pillar_i]
                points_array[:, index, :] = k_reduced_cp_array[:, pillar_j - jp, pillar_i - ip, jp, ip, :]
                index += 1
        # add extras for split pillars
        for pillar_j in range(grid.nj + 1):
            for pillar_i in range(grid.ni + 1):
                for e in range(extras_count[pillar_j, pillar_i]):
                    eli = extras_list_index[pillar_j, pillar_i] + e
                    (jp, ip) = extras_list[eli]
                    pillar_j_extra = pillar_j - jp
                    pillar_i_extra = pillar_i - ip
                    points_array[:, index, :] = k_reduced_cp_array[:, pillar_j_extra, pillar_i_extra, jp, ip, :]
                    index += 1
        assert index == (grid.nj + 1) * (grid.ni + 1) + len(extras_list)
    else:  # unsplit pillars
        points_array = np.zeros((grid.nk + 1, grid.nj + 1, grid.ni + 1, 3))
        for j in range(grid.nj + 1):
            for i in range(grid.ni + 1):
                (jp, ip) = primary_pillar_jip[j, i]
                points_array[:, j, i, :] = k_reduced_cp_array[:, j - jp, i - ip, jp, ip, :]
    grid.points_cached = points_array
    # add split pillar arrays to grid object
    if grid.has_split_coordinate_lines:
        _add_split_pillars_for_extras(grid, extras_count, extras_use, extras_list)


def _add_split_pillars_for_extras(grid, extras_count, extras_use, extras_list):
    log.debug('adding split pillar arrays to grid object')
    split_pillar_indices_list = []
    cumulative_length_list = []
    cols_for_extra_pillar_list = []
    cumulative_length = 0
    for pillar_j in range(grid.nj + 1):
        for pillar_i in range(grid.ni + 1):
            for e in range(extras_count[pillar_j, pillar_i]):
                split_pillar_indices_list.append(pillar_j * (grid.ni + 1) + pillar_i)
                use_count = 0
                for jp in range(2):
                    j = pillar_j - jp
                    if j < 0 or j >= grid.nj:
                        continue
                    for ip in range(2):
                        i = pillar_i - ip
                        if i < 0 or i >= grid.ni:
                            continue
                        if extras_use[j, i, jp, ip] == e:
                            use_count += 1
                            cols_for_extra_pillar_list.append((j * grid.ni) + i)
                assert (use_count > 0)
                cumulative_length += use_count
                cumulative_length_list.append(cumulative_length)
    log.debug('number of extra pillars: ' + str(len(split_pillar_indices_list)))
    assert (len(cumulative_length_list) == len(split_pillar_indices_list))
    grid.split_pillar_indices_cached = np.array(split_pillar_indices_list, dtype = 'int')
    log.debug('number of uses of extra pillars: ' + str(len(cols_for_extra_pillar_list)))
    assert (len(cols_for_extra_pillar_list) == np.count_nonzero(extras_use + 1))
    assert (len(cols_for_extra_pillar_list) == cumulative_length)
    grid.cols_for_split_pillars = np.array(cols_for_extra_pillar_list, dtype = 'int')
    assert (len(cumulative_length_list) == len(extras_list))
    grid.cols_for_split_pillars_cl = np.array(cumulative_length_list, dtype = 'int')
    grid.split_pillars_count = len(extras_list)


def _close_vertical_voids(grid, grid_cp, z_units):
    if grid.nk > 1:
        z_gap = grid_cp[1:, :, :, 0, :, :, :] - grid_cp[:-1, :, :, 1, :, :, :]
        max_gap = np.max(np.abs(z_gap))
        log.info('maximum vertical void distance after corner point interpolation: {0:.3f} {1}'.format(
            max_gap, z_units))
        # close vertical voids (includes shifting x & y)
        if max_gap > 0.0:
            log.debug('closing vertical voids')
            z_gap *= 0.5
            grid_cp[1:, :, :, 0, :, :, :] -= z_gap
            grid_cp[:-1, :, :, 1, :, :, :] += z_gap



"""RESQML GeologicUnitInterpretation class."""

# NB: in this module, the term 'unit' refers to a geological stratigraphic unit, i.e. a layer of rock, not a unit of measure

import logging

log = logging.getLogger(__name__)

import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
import resqpy.organize as rqo
import resqpy.strata
import resqpy.strata._strata_common as rqstc
from resqpy.olio.base import BaseResqpy
from resqpy.olio.xml_namespaces import curly_namespace as ns


class GeologicUnitInterpretation(BaseResqpy):
    """Class for RESQML Geologic Unit Interpretation objects.

    These objects can be parts in their own right. NB: Various more specialised classes also derive from this.

    RESQML documentation:

       The main class for data describing an opinion of a volume-based geologic feature or unit.
    """

    resqml_type = 'GeologicUnitInterpretation'

    def __init__(
            self,
            parent_model,
            uuid = None,
            title = None,
            domain = 'time',  # or should this be depth?
            geologic_unit_feature = None,
            composition = None,
            material_implacement = None,
            extra_metadata = None):
        """Initialises an geologic unit interpretation object.

        arguments:
           parent_model (model.Model): the model with which the new interpretation will be associated
           uuid (uuid.UUID, optional): the uuid of an existing RESQML geologic unit interpretation from which
              this object will be initialised
           title (str, optional): the citation title (feature name) of the new interpretation;
              ignored if uuid is not None
           domain (str, default 'time'): 'time', 'depth' or 'mixed', being the domain of the interpretation;
              ignored if uuid is not None
           geologic_unit_feature (organize.GeologicUnitFeature or StratigraphicUnitFeature, optional): the feature
              which this object is an interpretation of; ignored if uuid is not None
           composition (str, optional): the interpreted composition of the geologic unit; if present, must be
              in valid_compositions; ignored if uuid is not None
           material_implacement (str, optional): the interpeted material implacement of the geologic unit;
              if present, must be in valid_implacements, ie. 'autochtonous' or 'allochtonous';
              ignored if uuid is not None
           extra_metadata (dict, optional): extra metadata items for the new interpretation

        returns:
           a new geologic unit interpretation resqpy object which may be the basis of a derived class object

        note:
           the RESQML 2.0.1 schema definition includes a spurious trailing space in the names of two compositions;
           resqpy removes such spaces in the composition attribute as presented to calling code (but includes them
           in xml)
        """

        self.domain = domain
        self.geologic_unit_feature = geologic_unit_feature  # InterpretedFeature RESQML field
        self.has_occurred_during = (None, None)  # optional RESQML item
        if (not title) and geologic_unit_feature is not None:
            title = geologic_unit_feature.feature_name
        self.composition = composition  # optional RESQML item
        self.material_implacement = material_implacement  # optional RESQML item
        super().__init__(model = parent_model, uuid = uuid, title = title, extra_metadata = extra_metadata)
        if self.composition:
            assert self.composition in rqstc.valid_compositions,  \
               f'invalid composition {self.composition} for geological unit interpretation'
            self.composition = self.composition.strip()
        if self.material_implacement:
            assert self.material_implacement in rqstc.valid_implacements,  \
               f'invalid material implacement {self.material_implacement} for geological unit interpretation'

    def _load_from_xml(self):
        """Loads class specific attributes from xml for an existing RESQML object; called from BaseResqpy."""
        root_node = self.root
        assert root_node is not None
        self.domain = rqet.find_tag_text(root_node, 'Domain')
        # following allows derived StratigraphicUnitInterpretation to instantiate its own interpreted feature
        if self.resqml_type == 'GeologicUnitInterpretation':
            feature_uuid = bu.uuid_from_string(rqet.find_nested_tags_text(root_node, ['InterpretedFeature', 'UUID']))
            if feature_uuid is not None:
                self.geologic_unit_feature = rqo.GeologicUnitFeature(
                    self.model, uuid = feature_uuid, feature_name = self.model.title(uuid = feature_uuid))
        self.has_occurred_during = rqo.extract_has_occurred_during(root_node)
        self.composition = rqet.find_tag_text(root_node, 'GeologicUnitComposition')
        self.material_implacement = rqet.find_tag_text(root_node, 'GeologicUnitMaterialImplacement')

    def is_equivalent(self, other, check_extra_metadata = True):
        """Returns True if this interpretation is essentially the same as the other; otherwise False.

        arguments:
           other (GeologicUnitInterpretation or StratigraphicUnitInterpretation): the other interpretation to
              compare this one against
           check_extra_metadata (bool, default True): if True, then extra metadata items must match for the two
              interpretations to be deemed equivalent; if False, extra metadata is ignored in the comparison

        returns:
           bool: True if this interpretation is essentially the same as the other; False otherwise
        """

        # this method is coded to allow use by the derived StratigraphicUnitInterpretation class
        if other is None or not isinstance(other, type(self)):
            return False
        if self is other or bu.matching_uuids(self.uuid, other.uuid):
            return True
        if self.geologic_unit_feature is not None:
            if not self.geologic_unit_feature.is_equivalent(other.geologic_unit_feature):
                return False
        elif other.geologic_unit_feature is not None:
            return False
        if self.root is not None and other.root is not None:
            if rqet.citation_title_for_node(self.root) != rqet.citation_title_for_node(other.root):
                return False
        elif self.root is not None or other.root is not None:
            return False
        if check_extra_metadata and not rqo.equivalent_extra_metadata(self, other):
            return False
        return (self.composition == other.composition and self.material_implacement == other.material_implacement and
                self.domain == other.domain and
                rqo.equivalent_chrono_pairs(self.has_occurred_during, other.has_occurred_during))

    def create_xml(self, add_as_part = True, add_relationships = True, originator = None, reuse = True):
        """Creates a geologic unit interpretation xml tree.

        arguments:
           add_as_part (bool, default True): if True, the interpretation is added to the parent model as a high level part
           add_relationships (bool, default True): if True and add_as_part is True, a relationship is created with
              the referenced geologic unit feature
           originator (str, optional): if present, is used as the originator field of the citation block
           reuse (bool, default True): if True, the parent model is inspected for any equivalent interpretation and, if found,
              the uuid of this interpretation is set to that of the equivalent part

        returns:
           lxml.etree._Element: the root node of the newly created xml tree for the interpretation
        """

        # note: related feature xml must be created first and is referenced here
        # this method is coded to allow use by the derived StratigraphicUnitInterpretation class

        if reuse and self.try_reuse():
            return self.root
        gu = super().create_xml(add_as_part = False, originator = originator)

        assert self.geologic_unit_feature is not None
        guf_root = self.geologic_unit_feature.root
        assert guf_root is not None, 'interpreted feature not established for geologic unit interpretation'

        assert self.domain in rqstc.valid_domains, 'illegal domain value for geologic unit interpretation'
        dom_node = rqet.SubElement(gu, ns['resqml2'] + 'Domain')
        dom_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Domain')
        dom_node.text = self.domain

        self.model.create_ref_node('InterpretedFeature',
                                   self.geologic_unit_feature.title,
                                   self.geologic_unit_feature.uuid,
                                   content_type = self.geologic_unit_feature.resqml_type,
                                   root = gu)

        rqo.create_xml_has_occurred_during(self.model, gu, self.has_occurred_during)

        if self.composition is not None:
            assert self.composition in rqstc.valid_compositions,  \
                f'invalid composition {self.composition} for geologic unit interpretation'
            comp_node = rqet.SubElement(gu, ns['resqml2'] + 'GeologicUnitComposition')
            comp_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'GeologicUnitComposition')
            comp_node.text = self.composition
            if self.composition + ' ' in rqstc.valid_compositions:  # RESQML xsd has spurious trailing space for two compositions
                comp_node.text += ' '

        if self.material_implacement is not None:
            assert self.material_implacement in rqstc.valid_implacements,  \
                f'invalid material implacement {self.material_implacement} for geologic unit interpretation'
            mi_node = rqet.SubElement(gu, ns['resqml2'] + 'GeologicUnitMaterialImplacement')
            mi_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'GeologicUnitMaterialImplacement')
            mi_node.text = self.material_implacement

        if add_as_part:
            self.model.add_part('obj_' + self.resqml_type, self.uuid, gu)
            if add_relationships:
                self.model.create_reciprocal_relationship(gu, 'destinationObject', guf_root, 'sourceObject')

        return gu



"""Class for RESQML Boundary Feature Interpretation organizational objects."""

import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
import resqpy.organize
import resqpy.organize.boundary_feature as obf
import resqpy.organize._utils as ou
from resqpy.olio.base import BaseResqpy
from resqpy.olio.xml_namespaces import curly_namespace as ns


class BoundaryFeatureInterpretation(BaseResqpy):
    """Class for RESQML Horizon Interpretation organizational objects."""

    resqml_type = 'BoundaryFeatureInterpretation'
    valid_domains = ('depth', 'time', 'mixed')

    def __init__(self,
                 parent_model,
                 uuid = None,
                 title = None,
                 boundary_feature = None,
                 domain = 'depth',
                 extra_metadata = None):
        """Initialises a boundary feature interpretation organisational object."""

        # note: will create a paired BoundaryFeature object when loading from xml

        self.domain = domain
        self.boundary_feature = boundary_feature  # InterpretedFeature RESQML field, when not loading from xml
        self.feature_uuid = None if self.boundary_feature is None else self.boundary_feature.uuid
        if (not title) and self.boundary_feature is not None:
            title = self.boundary_feature.feature_name
        self.has_occurred_during = (None, None)
        super().__init__(model = parent_model, uuid = uuid, title = title, extra_metadata = extra_metadata)

    def _load_from_xml(self):
        root_node = self.root
        self.domain = rqet.find_tag_text(root_node, 'Domain')
        interp_feature_ref_node = rqet.find_tag(root_node, 'InterpretedFeature')
        assert interp_feature_ref_node is not None
        feature_root = self.model.referenced_node(interp_feature_ref_node)
        if feature_root is not None:
            self.feature_uuid = feature_root.attrib['uuid']
            self.boundary_feature = obf.BoundaryFeature(self.model,
                                                        uuid = self.feature_uuid,
                                                        feature_name = self.model.title_for_root(feature_root))
        self.has_occurred_during = ou.extract_has_occurred_during(root_node)

    def is_equivalent(self, other, check_extra_metadata = True):
        """Returns True if this interpretation is essentially the same as the other; otherwise False."""

        if other is None or not isinstance(other, BoundaryFeatureInterpretation):
            return False
        if self is other or bu.matching_uuids(self.uuid, other.uuid):
            return True
        if self.boundary_feature is not None:
            if not self.boundary_feature.is_equivalent(other.boundary_feature,
                                                       check_extra_metadata = check_extra_metadata):
                return False
        elif other.boundary_feature is not None:
            return False
        if self.title != other.title:
            return False
        if ((self.domain != other.domain) or
                not ou.equivalent_chrono_pairs(self.has_occurred_during, other.has_occurred_during)):
            return False
        if check_extra_metadata and not ou.equivalent_extra_metadata(self, other):
            return False
        return True

    def create_xml(self, add_as_part = True, add_relationships = True, originator = None, reuse = True):
        """Creates a boundary feature interpretation organisational xml tree."""

        # note: related boundary feature node should be created first and referenced here

        if not self.title:
            self.title = self.boundary_feature.feature_name

        if reuse and self.try_reuse():
            return self.root
        bfi = super().create_xml(add_as_part = False, originator = originator)

        if self.boundary_feature.root is None:
            bf_root = self.boundary_feature.create_xml(reuse = True)
            self.feature_uuid = bf_root.attrib['uuid']

        assert self.domain in self.valid_domains, 'illegal domain value for boundary feature interpretation'
        dom_node = rqet.SubElement(bfi, ns['resqml2'] + 'Domain')
        dom_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Domain')
        dom_node.text = self.domain

        ou.create_xml_has_occurred_during(self.model, bfi, self.has_occurred_during)

        self.model.create_ref_node('InterpretedFeature',
                                   self.model.title(uuid = self.feature_uuid),
                                   self.feature_uuid,
                                   content_type = 'obj_BoundaryFeature',
                                   root = bfi)

        if add_as_part:
            self.model.add_part('obj_BoundaryFeatureInterpretation', self.uuid, bfi)
            if add_relationships:
                self.model.create_reciprocal_relationship(bfi, 'destinationObject', self.boundary_feature.root,
                                                          'sourceObject')

        return bfi
